import numpy as np

from .typed import Dict, List


def convex_decomposition(mesh, **kwargs) -> List[Dict]:
    """
    Compute an approximate convex decomposition of a mesh.

    VHACD Parameters which can be passed as kwargs:

    Name                              Default
    -----------------------------------------
    maxConvexHulls                    64
    resolution                        400000
    minimumVolumePercentErrorAllowed  1.0
    maxRecursionDepth                 10
    shrinkWrap                        True
    fillMode                          "flood"
    maxNumVerticesPerCH               64
    asyncACD                          True
    minEdgeLength                     2
    findBestPlane                     False

    Parameters
    ----------
    mesh : trimesh.Trimesh
      Mesh to be decomposed into convex parts
    **kwargs : VHACD keyword arguments

    Returns
    -------
    mesh_args : list
      List of **kwargs for Trimeshes that are nearly
      convex and approximate the original.
    """
    from vhacdx import compute_vhacd

    # the faces are triangulated in a (len(face), ...vertex-index)
    # for vtkPolyData
    # i.e. so if shaped to four columns the first column is all 3
    faces = (
        np.column_stack((np.ones(len(mesh.faces), dtype=np.int64) * 3, mesh.faces))
        .ravel()
        .astype(np.uint32)
    )

    return [
        {"vertices": v, "faces": f}
        for v, f in compute_vhacd(mesh.vertices, faces, **kwargs)
    ]



try:
    from . import generic as g
except BaseException:
    import generic as g


class DecompositionTest(g.unittest.TestCase):
    def test_convex_decomposition(self):
        try:
            import vhacdx  # noqa
        except ImportError:
            return

        mesh = g.get_mesh("quadknot.obj")
        meshes = mesh.convex_decomposition()
        assert len(meshes) > 1
        for m in meshes:
            assert m.is_watertight
            assert m.is_convex


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


def explicit_laplacian_calculation(mesh, equal_weight=True, pinned_vertices=None):
    """
    Exact copy of non-sparse calculation of laplacian, for correctness testing.

    Calculate a sparse matrix for laplacian operations.
    Parameters
    -------------
    mesh : trimesh.Trimesh
      Input geometry
    equal_weight : bool
      If True, all neighbors will be considered equally
      If False, all neighbors will be weighted by inverse distance
    Returns
    ----------
    laplacian : scipy.sparse.coo.coo_matrix
      Laplacian operator
    """
    # get the vertex neighbors from the cache
    neighbors = mesh.vertex_neighbors

    # if a node is pinned, it will average his coordinates by himself
    # in practice it will not move
    if pinned_vertices is not None:
        for i in pinned_vertices:
            neighbors[i] = [i]

    # avoid hitting crc checks in loops
    vertices = mesh.vertices.view(g.np.ndarray)

    # stack neighbors to 1D arrays
    col = g.np.concatenate(neighbors)
    row = g.np.concatenate([[i] * len(n) for i, n in enumerate(neighbors)])

    if equal_weight:
        # equal weights for each neighbor
        data = g.np.concatenate([[1.0 / len(n)] * len(n) for n in neighbors])
    else:
        # umbrella weights, distance-weighted
        # use dot product of ones to replace array.sum(axis=1)
        ones = g.np.ones(3)
        # the distance from verticesex to neighbors
        norms = [
            1.0
            / g.np.maximum(
                1e-6, g.np.sqrt(g.np.dot((vertices[i] - vertices[n]) ** 2, ones))
            )
            for i, n in enumerate(neighbors)
        ]
        # normalize group and stack into single array
        data = g.np.concatenate([i / i.sum() for i in norms])

    # create the sparse matrix
    matrix = g.trimesh.graph.coo_matrix((data, (row, col)), shape=[len(vertices)] * 2)

    return matrix


class SmoothTest(g.unittest.TestCase):
    def test_laplacian_calculation(self):
        m = g.trimesh.creation.icosahedron()
        m.vertices, m.faces = g.trimesh.remesh.subdivide_to_size(m.vertices, m.faces, 0.1)

        explicit_laplacian = explicit_laplacian_calculation(m)
        laplacian = g.trimesh.smoothing.laplacian_calculation(m)

        assert g.np.allclose(explicit_laplacian.toarray(), laplacian.toarray())

        explicit_laplacian = explicit_laplacian_calculation(m, equal_weight=False)
        laplacian = g.trimesh.smoothing.laplacian_calculation(m, equal_weight=False)

        assert g.np.allclose(explicit_laplacian.toarray(), laplacian.toarray())

        explicit_laplacian = explicit_laplacian_calculation(m, pinned_vertices=[0, 1, 4])
        laplacian = g.trimesh.smoothing.laplacian_calculation(
            m, pinned_vertices=[0, 1, 4]
        )

        assert g.np.allclose(explicit_laplacian.toarray(), laplacian.toarray())

        explicit_laplacian = explicit_laplacian_calculation(
            m, equal_weight=False, pinned_vertices=[0, 1, 4]
        )
        laplacian = g.trimesh.smoothing.laplacian_calculation(
            m, equal_weight=False, pinned_vertices=[0, 1, 4]
        )

        assert g.np.allclose(explicit_laplacian.toarray(), laplacian.toarray())

    def test_smooth(self):
        """
        Load a collada scene with pycollada.
        """
        m = g.trimesh.creation.icosahedron()
        m.vertices, m.faces = g.trimesh.remesh.subdivide_to_size(m.vertices, m.faces, 0.1)

        s = m.copy()
        q = m.copy()
        f = m.copy()
        d = m.copy()
        b = m.copy()
        v = m.copy()

        assert m.is_volume

        # Equal Weights
        lap = g.trimesh.smoothing.laplacian_calculation(mesh=m, equal_weight=True)

        g.trimesh.smoothing.filter_laplacian(s, 0.5, 10, False, True, lap)
        g.trimesh.smoothing.filter_laplacian(q, 0.5, 10, True, True, lap)
        g.trimesh.smoothing.filter_humphrey(f, 0.1, 0.5, 10, lap)
        g.trimesh.smoothing.filter_taubin(d, 0.5, 0.53, 10, lap)
        g.trimesh.smoothing.filter_mut_dif_laplacian(b, 0.5, 10, False, lap)
        g.trimesh.smoothing.filter_mut_dif_laplacian(v, 0.5, 10, True, lap)

        assert s.is_volume
        assert q.is_volume
        assert f.is_volume
        assert d.is_volume
        assert b.is_volume
        assert v.is_volume

        assert g.np.isclose(s.volume, m.volume, rtol=0.1)
        assert g.np.isclose(q.volume, m.volume, rtol=0.1)
        assert g.np.isclose(f.volume, m.volume, rtol=0.1)
        assert g.np.isclose(d.volume, m.volume, rtol=0.1)
        assert g.np.isclose(b.volume, m.volume, rtol=0.1)
        assert g.np.isclose(v.volume, m.volume, rtol=0.1)

        s = m.copy()
        q = m.copy()
        f = m.copy()
        d = m.copy()
        b = m.copy()
        v = m.copy()

        # umbrella Weights
        lap = g.trimesh.smoothing.laplacian_calculation(m, equal_weight=False)

        g.trimesh.smoothing.filter_laplacian(s, 0.5, 10, False, True, lap)
        g.trimesh.smoothing.filter_laplacian(q, 0.5, 10, True, True, lap)
        g.trimesh.smoothing.filter_humphrey(f, 0.1, 0.5, 10, lap)
        g.trimesh.smoothing.filter_taubin(d, 0.5, 0.53, 10, lap)
        g.trimesh.smoothing.filter_mut_dif_laplacian(b, 0.5, 10, False, lap)
        g.trimesh.smoothing.filter_mut_dif_laplacian(v, 0.5, 10, True, lap)

        assert s.is_volume
        assert q.is_volume
        assert f.is_volume
        assert d.is_volume
        assert b.is_volume
        assert v.is_volume

        assert g.np.isclose(s.volume, m.volume, rtol=0.1)
        assert g.np.isclose(q.volume, m.volume, rtol=0.1)
        assert g.np.isclose(f.volume, m.volume, rtol=0.1)
        assert g.np.isclose(d.volume, m.volume, rtol=0.1)
        assert g.np.isclose(b.volume, m.volume, rtol=0.1)
        assert g.np.isclose(v.volume, m.volume, rtol=0.1)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


# tetToTris extracts outward facing triangles
# of a tetrahedron given in this order
#    1
#  / | \
# 0-----2
#  \ | /
#    3

triIdxs = g.np.ravel([[2, 1, 0], [3, 0, 1], [3, 2, 0], [3, 1, 2]])


def tetToTris(tet):
    return g.np.reshape(tet[triIdxs], (-1, 3))


def test_inertia_basic():
    t0 = g.np.array(
        [
            [-0.419575686853, -0.898655215203, -0.127965023308, 0.0],
            [0.712589964872, -0.413418145015, 0.566834172697, 0.0],
            [-0.562291548012, 0.146643245877, 0.813832890385, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    t1 = g.np.array(
        [
            [0.343159553585, 0.624765521319, -0.701362648103, 0.0],
            [0.509982849005, -0.750986657709, -0.419447891476, 0.0],
            [-0.788770571525, -0.213745370274, -0.57632794673, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    # make sure our transformations are actually still transformations
    assert g.np.abs(g.np.dot(t0, t0.T) - g.np.eye(4)).max() < 1e-10
    assert g.np.abs(g.np.dot(t1, t1.T) - g.np.eye(4)).max() < 1e-10

    c = g.trimesh.primitives.Cylinder(
        height=10,
        radius=1,
        sections=720,
        transform=t0,  # number of slices
    )
    c0m = c.moment_inertia.copy()
    c0 = g.trimesh.inertia.cylinder_inertia(
        c.volume, c.primitive.radius, c.primitive.height, c.primitive.transform
    )

    ct = g.np.abs((c0m / c0) - 1)

    # we are comparing an inertia tensor from a mesh of a cylinder
    # to an inertia tensor from an actual cylinder, so allow for some
    # discretization uncertainty
    assert ct.max() < 1e-3

    # check our principal axis calculation against this cylinder
    # the direction (long axis) of the cylinder should correspond to
    # the smallest principal component of inertia, AKA rotation along
    # the axis, rather than the other two which are perpendicular
    components, vectors = g.trimesh.inertia.principal_axis(c.moment_inertia)
    # inferred cylinder axis
    inferred = vectors[components.argmin()]

    # inferred cylinder axis should be close to actual cylinder axis
    axis_test = g.np.allclose(g.np.abs(inferred), g.np.abs(c.direction))
    assert axis_test

    # make sure Trimesh attribute is plumbed correctly
    assert g.np.allclose(c.principal_inertia_components, components)
    assert g.np.allclose(c.principal_inertia_vectors, vectors)

    # the other two axis of the cylinder should be identical
    assert g.np.abs(g.np.diff(g.np.sort(components)[-2:])).max() < 1e-8

    m = g.get_mesh("featuretype.STL")
    i0 = m.moment_inertia.copy()
    # rotate the moment of inertia
    i1 = g.trimesh.inertia.transform_inertia(transform=t0, inertia_tensor=i0)

    # rotate the mesh
    m.apply_transform(t0)
    # check to see if the rotated mesh + recomputed moment of inertia
    # is close to the rotated moment of inertia
    tf_test = g.np.abs((m.moment_inertia / i1) - 1)
    assert tf_test.max() < 1e-6

    # do it again with another transform
    i2 = g.trimesh.inertia.transform_inertia(transform=t1, inertia_tensor=i1)
    m.apply_transform(t1)
    tf_test = g.np.abs((m.moment_inertia / i2) - 1)
    assert tf_test.max() < 1e-6


def test_primitives():
    primitives = [
        g.trimesh.primitives.Cylinder(height=5),
        g.trimesh.primitives.Box(),
        g.trimesh.primitives.Sphere(radius=1.23),
    ]
    for p in primitives:
        for _i in range(100):
            # check to make sure the analytic inertia tensors are relatively
            # close to the meshed inertia tensor (order of magnitude and
            # sign)
            b = p.to_mesh()
            comparison = g.np.abs(p.moment_inertia - b.moment_inertia)
            c_max = comparison.max() / g.np.abs(p.moment_inertia).max()
            assert c_max < 0.1

            if hasattr(p.primitive, "transform"):
                matrix = g.trimesh.transformations.random_rotation_matrix()
                p.primitive.transform = matrix
            elif hasattr(p.primitive, "center"):
                p.primitive.center = g.random(3)


def test_tetrahedron():
    # Based on the 'numerical example' of the paper:

    # Explicit Exact Formulas for the 3-D Tetrahedron Inertia Tensor
    # in Terms of its Vertex Coordinates, [F. Tonon, 2004]
    # http://thescipub.com/pdf/jmssp.2005.8.11.pdf

    # set up given vertices
    vertices = g.np.float32(
        [
            [8.3322, -11.86875, 0.93355],
            [0.75523, 5.0, 16.37072],
            [52.61236, 5.0, -5.3858],
            [2.0, 5.0, 3.0],
        ]
    )

    # set up a simple trimesh tetrahedron
    tris = tetToTris(g.np.int32([0, 1, 2, 3]))
    tm_tet = g.trimesh.Trimesh(vertices, tris)

    # 'ground truth' values from the paper
    # however, there are some minor flaws

    # a small deviation in the last decimal of the mass-centers' x:
    CM_gt = g.np.float32([15.92492, 0.78281, 3.72962])

    # moment of inertia values
    a_mi, b_mi, c_mi = [43520.33257, 194711.28938, 191168.76173]
    # principle inertia values
    a_pi, b_pi, c_pi = [4417.6615, -46343.16662, 11996.20119]

    # NOTE: I guess there is a mistake in the paper
    # b' (Eq. 9e) computes from x and z values
    # c' (Eq. 9f) computes from x and y values
    # therefore the given matrix E_Q (Eq. 2) is not correct
    # b' and c' need to be swapped like this:
    MI_gt = g.np.float32(
        [[a_mi, -c_pi, -b_pi], [-c_pi, b_mi, -a_pi], [-b_pi, -a_pi, c_mi]]
    )

    # check center of mass
    assert g.np.allclose(CM_gt, tm_tet.center_mass)

    # check moment of inertia tensor
    assert g.np.allclose(MI_gt, tm_tet.moment_inertia)


def test_cube_with_tetras():
    # set up a unit cube, vertices in this order:
    #   1-----2
    #  /|    /|
    # 0-+---3 |
    # | 5---+-6
    # |/    |/
    # 4-----7

    vertices = (
        g.np.float32(
            [
                [-1, -1, 1],
                [-1, 1, 1],
                [1, 1, 1],
                [1, -1, 1],
                [-1, -1, -1],
                [-1, 1, -1],
                [1, 1, -1],
                [1, -1, -1],
            ]
        )
        * 0.5
    )

    # 6 quad faces for the cube
    quads = g.np.int32(
        [
            [3, 2, 1, 0],
            [0, 1, 5, 4],
            [1, 2, 6, 5],
            [2, 3, 7, 6],
            [3, 0, 4, 7],
            [4, 5, 6, 7],
        ]
    )

    # set up two different tetraherdalizations of a cube
    # using 5 tets (1 in the middle and 4 around)
    tetsA = g.np.int32(
        [[0, 1, 3, 4], [1, 2, 3, 6], [1, 4, 5, 6], [3, 4, 6, 7], [1, 3, 4, 6]]
    )
    # using 6 sets (around the cube diagonal)
    tetsB = g.np.int32(
        [
            [0, 1, 2, 6],
            [0, 1, 6, 5],
            [0, 4, 5, 6],
            [0, 4, 6, 7],
            [0, 3, 7, 6],
            [0, 2, 3, 6],
        ]
    )

    # create a trimesh cube from vertices and faces
    tm_cube = g.trimesh.Trimesh(vertices, quads)

    # https://en.wikipedia.org/wiki/List_of_moments_of_inertia
    # ground truth for a unit cube with side length s = mass m = 1
    # I = 1/6 * m * s^2
    MI_gt = g.np.eye(3) / 6
    assert g.np.allclose(MI_gt, tm_cube.moment_inertia)

    # compare both tetrahedralizations
    # should be equivalent to each other and to the cube
    for tets in [tetsA, tetsB]:
        # create trimesh tets from vertices and triangles
        tm_tets = [g.trimesh.Trimesh(vertices, tetToTris(t)) for t in tets]

        # get mass, center of mass, and moment of inertia for each tet
        Ms = [tm_tet.mass for tm_tet in tm_tets]
        CMs = [tm_tet.center_mass for tm_tet in tm_tets]
        MIs = [tm_tet.moment_inertia for tm_tet in tm_tets]

        # compute total mass and center of mass
        mass = g.np.sum(Ms)
        center_mass = g.np.dot(Ms, CMs)

        # compare with unit cube
        assert g.np.abs(tm_cube.mass - mass) < 1e-6
        assert g.np.allclose(tm_cube.center_mass, center_mass)

        # the moment of inertia tensors for the individual tetrahedra
        # have to be re-assembled, using the parallel-axis-theorem
        # https://en.wikipedia.org/wiki/Parallel_axis_theorem#Tensor_generalization

        E = g.np.eye(3)
        MI = g.np.zeros((3, 3), g.np.float32)
        for i in range(len(tm_tets)):
            R = CMs[i] - center_mass
            MI += MIs[i] + Ms[i] * (g.np.dot(R, R) * E - g.np.outer(R, R))

        assert g.np.allclose(MI_gt, MI)


def test_frame_inertia_box():
    # set up a rectangular cuboid, vertices as visible below.
    # The edges have the lengths 1,2 and 3
    #           ^ x0, x2
    #    y2 ^  /    b
    #       | 1-----------2
    #       |/|          /|
    #       0-+---------3-+----> y0, x1, z2
    #      /| |         | |
    #     / | |         | |
    # y1 v  | |         | |h
    #       | |         | |
    #       | |         | |
    #       | 5---------+-6
    #       |/          |/ d
    #       4-----------7
    #       |
    #       v z0, z1

    h = 3
    b = 2
    d = 1
    # definition in frame 0
    vertices = g.np.float32(
        [
            [0, 0, 0],
            [d, 0, 0],
            [d, b, 0],
            [0, b, 0],
            [0, 0, h],
            [d, 0, h],
            [d, b, h],
            [0, b, h],
        ]
    )

    # 6 quad faces for the cube
    quads = g.np.int32(
        [
            [3, 2, 1, 0],
            [0, 1, 5, 4],
            [1, 2, 6, 5],
            [2, 3, 7, 6],
            [3, 0, 4, 7],
            [4, 5, 6, 7],
        ]
    )

    # create a trimesh cube from vertices and faces
    tm_cube = g.trimesh.Trimesh(vertices, quads)

    # set density of 1 and calculate mass
    tm_cube.density = 1
    mass = tm_cube.density * h * b * d
    assert tm_cube.mass == mass

    def parallel_axis_theorem(inertia, mass, a1, a2, a3):
        """
        Apply parallel axis theorem

        Parameters
        ------------
        inertia: (3,3) float
          inertia tensor around centre of mass
        mass : float
        a1, a2, a3: float
          difference between desired rotation center and center of mass in x,y,z

        Returns
        -----------
        rotated_inertia : (3,3) float
        Inertia tensor

        """
        # copy from wikipedia
        return inertia + mass * g.np.array(
            [
                [a2**2 + a3**2, -a1 * a2, -a1 * a3],
                [-a1 * a2, a1**2 + a3**2, -a2 * a3],
                [-a1 * a3, -a2 * a3, a1**2 + a2**2],
            ]
        )

    # CHECK FRAME 0
    # analytical calculations of inertia tensor by hand
    inertia0 = 0.083333333333 * mass * g.np.diag([h**2 + b**2, h**2 + d**2, b**2 + d**2])
    a1 = -0.5 * d
    a2 = -0.5 * b
    a3 = -0.5 * h
    inertia0 = parallel_axis_theorem(inertia0, mass, a1, a2, a3)
    # transformation from mesh base frame to frame 0
    t0 = g.np.eye(4)
    assert g.np.allclose(tm_cube.moment_inertia_frame(t0), inertia0)

    # CHECK FRAME 1
    # analytical calculations of inertia tensor by hand
    inertia1 = 0.083333333333 * mass * g.np.diag([h**2 + d**2, h**2 + b**2, b**2 + d**2])
    a1 = -0.5 * b
    a2 = 0.5 * d
    a3 = -0.5 * h
    inertia1 = parallel_axis_theorem(inertia1, mass, a1, a2, a3)
    # transformation from mesh base frame to frame 1
    # rotation of 90 deg around z-Axis
    t1 = g.trimesh.transformations.rotation_matrix(g.np.pi * 0.5, [0, 0, 1], [0, 0, 0])
    assert g.np.allclose(tm_cube.moment_inertia_frame(t1), inertia1)

    # CHECK FRAME 2
    # analytical calculations of inertia tensor by hand
    inertia2 = 0.083333333333 * mass * g.np.diag([h**2 + b**2, b**2 + d**2, h**2 + d**2])
    a1 = -0.5 * d
    a2 = 0.5 * h
    a3 = -0.5 * b
    inertia2 = parallel_axis_theorem(inertia2, mass, a1, a2, a3)
    # transformation from mesh base frame to frame 2
    # rotation of -90 deg around x-Axis
    t2 = g.trimesh.transformations.rotation_matrix(-g.np.pi * 0.5, [1, 0, 0], [0, 0, 0])
    assert g.np.allclose(tm_cube.moment_inertia_frame(t2), inertia2)


def test_scene():
    # check our transforming of inertia matrices using the
    # parallel axis theorum by checking the inertia of a scene
    # with multiple instances transformed all over the scene
    # in two ways: by dumping it into a single mesh and then
    # running the heavily validated surface integral method,
    # and also running our "transform every matrix" logic
    # contained in `trimesh.inertia.scene_inertia`

    # scene with instancing
    s = g.get_mesh("cycloidal.3DXML")
    s._cache.clear()

    with g.Profiler() as P:
        ms = s.to_mesh()
        total_dump = ms.moment_inertia
    g.log.debug(P.output_text())

    # while we're at it check to make sure our composed
    # center mass, volume, and area calcs are all working
    assert g.np.allclose(s.center_mass, ms.center_mass)
    # dumped inertia should be the same as transformed inertia
    assert g.np.allclose(s.moment_inertia, ms.moment_inertia)
    assert g.np.allclose(s.volume, ms.volume)
    assert g.np.allclose(s.area, ms.area)

    # that produced the moment of inertia around the center of mass
    # of the scene, so for the `moment_inertia_frame` calculation
    # compare against that
    transform = g.tf.translation_matrix(ms.center_mass)
    s._cache.clear()

    with g.Profiler() as P:
        total_scene = g.trimesh.inertia.scene_inertia(s, transform=transform)
    g.log.debug(P.output_text())

    # compare the two calculation methods by percent
    diff = g.np.abs((total_dump - total_scene) / total_dump)
    assert diff.shape == (3, 3)

    # the two methods should return essentially identical results
    assert g.np.abs(diff).max() < 1e-3


def test_points_inertia():
    # function we're testing
    inertia = g.trimesh.inertia.points_inertia

    # should be m * r**2 around XY, zero in Z
    t = inertia([[0, 0, 10]], weights=1.0, at_center_mass=False)
    assert g.np.allclose(t, [[100.0, 0.0, 0.0], [0.0, 100.0, 0.0], [0.0, 0.0, 0.0]])

    # run with every point assigned a weight of `1.0`
    # by default should offset to the center of mass
    t = inertia(g.np.eye(3) + 12342.234234, weights=1.0)
    assert g.np.allclose(
        t,
        [
            [1.33333333, 0.33333333, 0.33333333],
            [0.33333333, 1.33333333, 0.33333333],
            [0.33333333, 0.33333333, 1.33333333],
        ],
    )

    # check a few masses and radii
    for mass in [0.5, 1.0, 12.12123]:
        for radius in [0.00123, 10.0, 123.1232785]:
            # a point along each of the XYZ axis
            points = g.np.eye(3) * radius
            collect = []

            for axis in range(3):
                # weight just one point
                weights = g.np.zeros(3)
                weights[axis] = mass

                calc = inertia(points, weights=weights, at_center_mass=False)

                # should be mr^2 for every axis except our chosen one
                expected = g.np.eye(3) * mass * radius**2
                expected[axis][axis] = 0.0
                assert g.np.allclose(expected, calc)

                # collect to compare summed result
                collect.append(calc)

            # try it with every point weighted
            summed = inertia(
                points, weights=g.np.full(len(points), mass), at_center_mass=False
            )
            # should match the sum of the individual options
            assert g.np.allclose(summed, g.np.sum(collect, axis=0))


def test_points_inertia_sample():
    # check to see that randomly sampled volume of a box
    # moved somewhere in space is "about equal" with the
    # approximate inertia of the points compared to the
    # exact intertia tensor of the box primitive

    # start with a random but deterministic rotation matrix
    matrix = g.np.array(
        [
            [-0.4984, 0.20643, 0.84201, -377.39084],
            [-0.3338, -0.94205, 0.03338, -348.96965],
            [0.80011, -0.26442, 0.53843, 494.61203],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    box = g.trimesh.primitives.Box(transform=matrix)

    # random points inside the box volume
    samples = box.sample_volume(10000)

    # calculate the inertia tensor of these points
    check = g.trimesh.inertia.points_inertia(samples, weights=box.mass / len(samples))

    # compare to the exact inertia tensor of the box primitive
    truth = box.moment_inertia

    # should be "about the same" but this will still be noisy
    # but should asymptote to exactly equal as sample count goes up
    assert g.np.allclose(check, truth, atol=0.01, rtol=0.01)


def test_mass_against_solidworks():
    # inertia numbers pulled from solidworks
    for truth in g.data["mass_properties"]:
        # load the mesh
        mesh = g.get_mesh(truth["filename"])
        # calculate our inertia
        calc = g.trimesh.triangles.mass_properties(
            triangles=mesh.triangles, density=truth["density"], skip_inertia=False
        )
        # compare values we have
        for key in truth.keys():
            if key == "area":
                assert g.np.isclose(mesh.area, truth[key], atol=1e-2)
            elif key in dir(calc):
                assert g.np.allclose(calc[key], truth[key], atol=1e-2)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    test_mass_against_solidworks()



try:
    from . import generic as g
except BaseException:
    import generic as g


def test_imc_sphere():
    for radius in [0.1, 1.0, 3.1459, 29.20]:
        m = g.trimesh.creation.icosphere(subdivisions=4, radius=radius)
        # the analytical value for a sphere
        reference = 4 * g.np.pi * radius
        assert g.np.isclose(m.integral_mean_curvature, reference, rtol=1e-3)


def test_imc_capsule():
    radius = 1.2
    for aspect_ratio in [0.0, 0.5, 1.0, 4.0, 100]:
        L = aspect_ratio * radius
        m = g.trimesh.creation.capsule(height=L, radius=radius, count=[64, 64])
        # the analytical value for a capsule
        reference = g.np.pi * (L + 4 * radius)
        assert g.np.isclose(m.integral_mean_curvature, reference, rtol=1e-3)


def test_imc_box(iterations: int = 4):
    extents = 1 - g.random((iterations, 3))
    for extent in extents:
        m = g.trimesh.creation.box(extents=extent)
        # only the right angle (=pi/2) edges contribute,
        # edges that lie on the faces of the box do not
        reference = 4 * extent.sum() * g.np.pi / 2 * 0.5
        assert g.np.isclose(m.integral_mean_curvature, reference, rtol=1e-3)


def test_imc_torus():
    r = 1.1
    for aspect_ratio in [1.1, 2.1, 4.0, 10.0]:
        R = r * aspect_ratio
        m = g.trimesh.creation.torus(
            major_radius=R, minor_radius=r, major_sections=100, minor_sections=100
        )
        reference = 2 * (g.np.pi**2) * R
        assert g.np.isclose(m.integral_mean_curvature, reference, rtol=1e-2)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()

    from pyinstrument import Profiler

    with Profiler() as P:
        test_imc_sphere()
        test_imc_box()
        test_imc_torus()
        test_imc_capsule()
    P.print()



# Copied from
# https://github.com/mmatl/pyrender/blob/master/pyrender/trackball.py

# MIT License
#
# Copyright (c) 2019 Matthew Matl
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Trackball class for 3D manipulation of viewpoints."""

import numpy as np

from .. import transformations


class Trackball:
    """A trackball class for creating camera transforms from mouse movements."""

    STATE_ROTATE = 0
    STATE_PAN = 1
    STATE_ROLL = 2
    STATE_ZOOM = 3

    def __init__(self, pose, size, scale, target=None):
        """Initialize a trackball with an initial camera-to-world pose
        and the given parameters.

        Parameters
        ----------
        pose : [4,4]
            An initial camera-to-world pose for the trackball.

        size : (float, float)
            The width and height of the camera image in pixels.

        scale : float
            The diagonal of the scene's bounding box --
            used for ensuring translation motions are sufficiently
            fast for differently-sized scenes.

        target : (3,) float
            The center of the scene in world coordinates.
            The trackball will revolve around this point.
        """
        self._size = np.array(size)
        self._scale = float(scale)

        self._pose = pose
        self._n_pose = pose

        if target is None:
            self._target = np.array([0.0, 0.0, 0.0])
            self._n_target = np.array([0.0, 0.0, 0.0])
        else:
            self._target = target
            self._n_target = target

        self._state = Trackball.STATE_ROTATE

    @property
    def pose(self):
        """autolab_core.RigidTransform : The current camera-to-world pose."""
        return self._n_pose

    def set_state(self, state):
        """Set the state of the trackball in order to change the effect of
        dragging motions.

        Parameters
        ----------
        state : int
            One of Trackball.STATE_ROTATE, Trackball.STATE_PAN,
            Trackball.STATE_ROLL, and Trackball.STATE_ZOOM.
        """
        self._state = state

    def resize(self, size):
        """Resize the window.

        Parameters
        ----------
        size : (float, float)
            The new width and height of the camera image in pixels.
        """
        self._size = np.array(size)

    def down(self, point):
        """Record an initial mouse press at a given point.

        Parameters
        ----------
        point : (2,) int
            The x and y pixel coordinates of the mouse press.
        """
        self._pdown = np.array(point, dtype=np.float32)
        self._pose = self._n_pose
        self._target = self._n_target

    def drag(self, point):
        """Update the tracball during a drag.

        Parameters
        ----------
        point : (2,) int
            The current x and y pixel coordinates of the mouse during a drag.
            This will compute a movement for the trackball with the relative
            motion between this point and the one marked by down().
        """
        point = np.array(point, dtype=np.float32)
        # get the "down" point defaulting to current point making
        # this a no-op if the "down" event didn't trigger for some reason
        dx, dy = point - getattr(self, "_pdown", point)
        mindim = 0.3 * np.min(self._size)

        target = self._target
        x_axis = self._pose[:3, 0].flatten()
        y_axis = self._pose[:3, 1].flatten()
        z_axis = self._pose[:3, 2].flatten()
        eye = self._pose[:3, 3].flatten()

        # Interpret drag as a rotation
        if self._state == Trackball.STATE_ROTATE:
            x_angle = -dx / mindim
            x_rot_mat = transformations.rotation_matrix(x_angle, y_axis, target)

            y_angle = dy / mindim
            y_rot_mat = transformations.rotation_matrix(y_angle, x_axis, target)

            self._n_pose = y_rot_mat.dot(x_rot_mat.dot(self._pose))

        # Interpret drag as a roll about the camera axis
        elif self._state == Trackball.STATE_ROLL:
            center = self._size / 2.0
            v_init = self._pdown - center
            v_curr = point - center
            v_init = v_init / np.linalg.norm(v_init)
            v_curr = v_curr / np.linalg.norm(v_curr)

            theta = -np.arctan2(v_curr[1], v_curr[0]) + np.arctan2(v_init[1], v_init[0])

            rot_mat = transformations.rotation_matrix(theta, z_axis, target)

            self._n_pose = rot_mat.dot(self._pose)

        # Interpret drag as a camera pan in view plane
        elif self._state == Trackball.STATE_PAN:
            dx = -dx / (5.0 * mindim) * self._scale
            dy = -dy / (5.0 * mindim) * self._scale

            translation = dx * x_axis + dy * y_axis
            self._n_target = self._target + translation
            t_tf = np.eye(4)
            t_tf[:3, 3] = translation
            self._n_pose = t_tf.dot(self._pose)

        # Interpret drag as a zoom motion
        elif self._state == Trackball.STATE_ZOOM:
            radius = np.linalg.norm(eye - target)
            ratio = 0.0
            if dy > 0:
                ratio = np.exp(abs(dy) / (0.5 * self._size[1])) - 1.0
            elif dy < 0:
                ratio = 1.0 - np.exp(dy / (0.5 * (self._size[1])))
            translation = -np.sign(dy) * ratio * radius * z_axis
            t_tf = np.eye(4)
            t_tf[:3, 3] = translation
            self._n_pose = t_tf.dot(self._pose)

    def scroll(self, clicks):
        """Zoom using a mouse scroll wheel motion.

        Parameters
        ----------
        clicks : int
            The number of clicks. Positive numbers indicate forward wheel
            movement.
        """
        target = self._target
        ratio = 0.90

        mult = 1.0
        if clicks > 0:
            mult = ratio**clicks
        elif clicks < 0:
            mult = (1.0 / ratio) ** abs(clicks)

        z_axis = self._n_pose[:3, 2].flatten()
        eye = self._n_pose[:3, 3].flatten()
        radius = np.linalg.norm(eye - target)
        translation = (mult * radius - radius) * z_axis
        t_tf = np.eye(4)
        t_tf[:3, 3] = translation
        self._n_pose = t_tf.dot(self._n_pose)

        z_axis = self._pose[:3, 2].flatten()
        eye = self._pose[:3, 3].flatten()
        radius = np.linalg.norm(eye - target)
        translation = (mult * radius - radius) * z_axis
        t_tf = np.eye(4)
        t_tf[:3, 3] = translation
        self._pose = t_tf.dot(self._pose)

    def rotate(self, azimuth, axis=None):
        """Rotate the trackball about the "Up" axis by azimuth radians.

        Parameters
        ----------
        azimuth : float
            The number of radians to rotate.
        """
        target = self._target

        y_axis = self._n_pose[:3, 1].flatten()
        if axis is not None:
            y_axis = axis
        x_rot_mat = transformations.rotation_matrix(azimuth, y_axis, target)
        self._n_pose = x_rot_mat.dot(self._n_pose)

        y_axis = self._pose[:3, 1].flatten()
        if axis is not None:
            y_axis = axis
        x_rot_mat = transformations.rotation_matrix(azimuth, y_axis, target)
        self._pose = x_rot_mat.dot(self._pose)



"""
comparison.py
----------------

Provide methods for quickly hashing and comparing meshes.
"""

from hashlib import sha256

import numpy as np

from . import util
from .constants import tol

# how many significant figures to use for each
# field of the identifier based on hand-tuning
id_sigfig = np.array(
    [
        5,  # area
        10,  # euler number
        5,  # area/volume ratio
        2,  # convex/mesh area ratio
        2,  # convex area/volume ratio
        3,  # max radius squared / area
        1,
    ]
)  # sign of triangle count for mirrored


def identifier_simple(mesh):
    """
    Return a basic identifier for a mesh consisting of
    properties that have been hand tuned to be somewhat
    robust to rigid transformations and different
    tessellations.

    Parameters
    ------------
    mesh : trimesh.Trimesh
      Source geometry

    Returns
    ----------
    identifier : (7,) float
      Identifying values of the mesh
    """
    # verify the cache once
    mesh._cache.verify()

    # don't check hashes during identifier as we aren't
    # changing any data values of the mesh inside block
    # if we did change values in cache block things would break
    with mesh._cache:
        # pre-allocate identifier so indexes of values can't move around
        # like they might if we used hstack or something else
        identifier = np.zeros(7, dtype=np.float64)
        # avoid thrashing the cache unnecessarily
        mesh_area = mesh.area
        # start with properties that are valid regardless of watertightness
        # note that we're going to try to make all parameters relative
        # to area so other values don't get blown up at weird scales
        identifier[0] = mesh_area
        # avoid divide-by-zero later
        if mesh_area < tol.merge:
            mesh_area = 1.0
        # topological constant and the only thing we can really
        # trust in this fallen world
        identifier[1] = mesh.euler_number

        # if we have a watertight mesh include volume and inertia
        if mesh.is_volume:
            # side length of a cube ratio
            # 1.0 for cubes, different values for other things
            identifier[2] = ((mesh_area / 6.0) ** (1.0 / 2.0)) / (
                mesh.volume ** (1.0 / 3.0)
            )
        else:
            # if we don't have a watertight mesh add information about the
            # convex hull which is slow to compute and unreliable
            try:
                # get the hull area and volume
                hull = mesh.convex_hull
                hull_area = hull.area
                hull_volume = hull.volume
            except BaseException:
                # in-plane or single point geometry has no hull
                hull_area = 6.0
                hull_volume = 1.0
            # just what we're looking for in a hash but hey
            identifier[3] = mesh_area / hull_area
            # cube side length ratio for the hull
            if hull_volume > 1e-12:
                identifier[4] = ((hull_area / 6.0) ** (1.0 / 2.0)) / (
                    hull_volume ** (1.0 / 3.0)
                )
            # calculate maximum mesh radius
            vertices = mesh.vertices - mesh.centroid
            # add in max radius^2 to area ratio
            R2 = np.dot((vertices**2), [1, 1, 1]).max()
            identifier[5] = R2 / mesh_area

    # mirrored meshes will look identical in terms of
    # area, volume, etc: use a count of relative edge
    # lengths to differentiate identical but mirrored meshes
    # this doesn't work well on meshes with a small number of faces
    if len(mesh.faces) > 50:
        # does this mesh have edges that differ substantially in length
        # if not this method for detecting reflection will not work
        # and the result will definitely be garbage
        edges_length = mesh.edges_unique_length
        variance = edges_length.std() / edges_length.mean()
        if variance > 0.25:
            # the length of each edge in faces
            norms = edges_length[mesh.edges_unique_inverse].reshape((-1, 3))
            # stack edge length and get the relative difference
            stack = np.diff(np.column_stack((norms, norms[:, 0])), axis=1)
            pick_idx = np.abs(stack).argmin(axis=1)
            # get the edge length diff
            pick = stack.reshape(-1)[pick_idx + (np.arange(len(pick_idx)) * 3)]
            # reduce to the bare minimum that tests stable
            identifier[6] = np.sign(pick.sum())
    return identifier


def identifier_hash(identifier):
    """
    Hash an identifier array in a way that is hand-tuned to be
    somewhat robust to likely changes.

    Parameters
    ------------
    identifier : (n,) float
      Vector of properties

    Returns
    ----------
    hash : (64,) str
      A SHA256 of the identifier vector at hand-tuned precision.
    """

    # convert identifier to integers and order of magnitude
    as_int, multiplier = util.sigfig_int(identifier, id_sigfig)

    # make all scales positive
    if (multiplier < 0).any():
        multiplier += np.abs(multiplier.min())
    data = (as_int * (10**multiplier)).astype(np.int64)
    return sha256(data.tobytes()).hexdigest()



"""
viewer
-------------

View meshes and scenes via pyglet or inline HTML.
"""

from .. import exceptions
from .notebook import (
    in_notebook,
    scene_to_html,
    scene_to_mo_notebook,
    scene_to_notebook,
)

try:
    # try importing windowed which will fail
    # if we can't create an openGL context
    from .windowed import SceneViewer, render_scene
except BaseException as E:
    # if windowed failed to import only raise
    # the exception if someone tries to use them
    SceneViewer = exceptions.ExceptionWrapper(E)
    render_scene = exceptions.ExceptionWrapper(E)


try:
    from .widget import SceneWidget
except BaseException as E:
    SceneWidget = exceptions.ExceptionWrapper(E)


# this is only standard library imports

# explicitly list imports in __all__
# as otherwise flake8 gets mad
__all__ = [
    "SceneViewer",
    "SceneWidget",
    "in_notebook",
    "render_scene",
    "scene_to_html",
    "scene_to_mo_notebook",
    "scene_to_notebook",
]



try:
    from . import generic as g
except BaseException:
    import generic as g


class LoaderTest(g.unittest.TestCase):
    def test_remote(self):
        """
        Try loading a remote mesh using requests
        """
        # get a unit cube from localhost
        with g.serve_meshes() as address:
            scene = g.trimesh.exchange.load.load_remote(url=address + "/unit_cube.STL")

        mesh = scene.to_mesh()
        assert g.np.isclose(mesh.volume, 1.0)
        assert isinstance(mesh, g.trimesh.Trimesh)

    def test_stl(self):
        model = g.get_mesh("empty.stl")
        assert model.is_empty

    def test_meshio(self):
        try:
            import meshio  # NOQA
        except BaseException:
            return
        # if meshio is importable we should be able to load this
        m = g.get_mesh("insulated.msh")
        assert len(m.faces) > 0
        assert m.area > 1e-5

    def test_load_mesh(self):
        # test the influence of the parameter `process` for `load_mesh`
        with open(g.os.path.join(g.dir_models, "featuretype.STL"), "rb") as file_obj:
            mesh = g.trimesh.load_mesh(file_obj=file_obj, file_type="stl", process=False)
            # check that number of vertices is not being reduced
            assert len(mesh.vertices) == 10428

        with open(g.os.path.join(g.dir_models, "featuretype.STL"), "rb") as file_obj:
            mesh = g.trimesh.load_mesh(file_obj=file_obj, file_type="stl", process=True)
            # check that number of vertices is being reduced
            assert len(mesh.vertices) == 1722

    def test_fileobj(self):
        # make sure we don't close file objects that were passed
        # check load_mesh
        file_obj = open(g.os.path.join(g.dir_models, "featuretype.STL"), "rb")
        assert not file_obj.closed
        mesh = g.trimesh.load_mesh(file_obj=file_obj, file_type="stl")
        # should have actually loaded the mesh
        assert len(mesh.faces) == 3476
        # should not close the file object
        assert not file_obj.closed
        # clean up
        file_obj.close()

        # check load_path
        file_obj = open(g.os.path.join(g.dir_models, "2D", "wrench.dxf"), "rb")
        assert not file_obj.closed
        path = g.trimesh.load(file_obj=file_obj, file_type="dxf")
        assert g.np.isclose(path.area, 1.667, atol=1e-2)
        # should have actually loaded the path
        # should not close the file object
        assert not file_obj.closed
        # clean up
        file_obj.close()

    def test_load_bz2(self):
        mesh = g.trimesh.load(g.os.path.join(g.dir_models, "rock.obj.bz2"), force="mesh")
        assert mesh.is_volume


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
Load all the meshes we can get our hands on and check things, stuff.
"""

try:
    from . import generic as g
except BaseException:
    import generic as g


class GeomTests(g.unittest.TestCase):
    def test_triangulate(self):
        from trimesh.geometry import triangulate_quads as tq

        # create some triangles and quads
        tri = (g.random((100, 3)) * 100).astype(g.np.int64)
        quad = (g.random((100, 4)) * 100).astype(g.np.int64)

        # should just exit early for triangles
        assert g.np.allclose(tri, tq(tri.tolist()))

        # should produce two triangles for each quad
        assert len(tq(quad)) == 2 * len(quad)

        # create a mixed list of triangles and quads
        mixed = tri.tolist()
        mixed.extend(quad.tolist())
        result = tq(mixed)

        # make sure the result has the right number of elements
        assert result.shape == (len(tri) + len(quad) * 2, 3)

        # called on empty arrays should be empty
        assert len(tq([])) == 0


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class PolygonTests(g.unittest.TestCase):
    def test_edges(self):
        """
        Test edges_to_polygon
        """
        m = g.get_mesh("featuretype.STL")

        # get a polygon for the second largest facet
        index = m.facets_area.argsort()[-2]
        normal = m.facets_normal[index]
        origin = m._cache["facets_origin"][index]
        T = g.trimesh.geometry.plane_transform(origin, normal)
        vertices = g.trimesh.transform_points(m.vertices, T)[:, :2]

        # find boundary edges for the facet
        edges = m.edges_sorted.reshape((-1, 6))[m.facets[index]].reshape((-1, 2))
        group = g.trimesh.grouping.group_rows(edges, require_count=1)

        # run the polygon conversion
        polygon = g.trimesh.path.polygons.edges_to_polygons(
            edges=edges[group], vertices=vertices
        )

        assert len(polygon) == 1
        assert g.np.isclose(polygon[0].area, m.facets_area[index])

        # try transforming the polygon around
        M = g.np.eye(3)
        M[0][2] = 10.0
        P2 = g.trimesh.path.polygons.transform_polygon(polygon[0], M)
        distance = (
            g.np.array(P2.centroid.coords)[0] - g.np.array(polygon[0].centroid.coords)[0]
        )
        assert g.np.allclose(distance, [10.0, 0])

    def test_random_polygon(self):
        """
        Test creation of random polygons
        """
        p = g.trimesh.path.polygons.random_polygon()
        assert p.area > 0.0
        assert p.is_valid

    def test_sample(self):
        """
        Test random sampling of polygons
        """

        p = g.Point([0, 0]).buffer(1.0)
        count = 100

        s = g.trimesh.path.polygons.sample(p, count=count)
        assert len(s) <= count
        assert s.shape[1] == 2

        radius = (s**2).sum(axis=1).max()
        assert radius < (1.0 + 1e-8)

        # test Path2D sample wiring
        path = g.trimesh.load_path(p)
        s = path.sample(count=count)
        assert len(s) <= count
        assert s.shape[1] == 2
        radius = (s**2).sum(axis=1).max()
        assert radius < (1.0 + 1e-8)

        # try getting OBB of samples
        _T, extents = g.trimesh.path.polygons.polygon_obb(s)
        # OBB of samples should be less than diameter of circle
        diameter = g.np.ptp(g.np.reshape(p.bounds, (2, 2)), axis=0).max()
        assert (extents <= diameter).all()

        # test sampling with multiple bodies
        for i in range(3):
            assert g.np.isclose(path.area, p.area * (i + 1))
            path = path + g.trimesh.load_path(g.Point([(i + 2) * 2, 0]).buffer(1.0))
            s = path.sample(count=count)
            assert s.shape[1] == 2

    def test_project(self):
        m = g.trimesh.creation.icosphere(subdivisions=4)

        p = [g.trimesh.path.polygons.projected(m, normal=n) for n in g.random((100, 3))]

        # sphere projection should never have interiors
        assert all(len(i.interiors) == 0 for i in p)
        # sphere projected area should always be close to pi
        assert g.np.allclose([i.area for i in p], g.np.pi, atol=0.05)

    def test_project_backface(self):
        m = g.trimesh.Trimesh(
            vertices=[[0, 0, 0], [0, 1, 0], [1, 0, 0]], faces=[[0, 1, 2]]
        )

        # check ignore_sign argument
        front = m.projected(m.face_normals[0], ignore_sign=False)
        assert len(front.entities) == 1
        front = m.projected(m.face_normals[0], ignore_sign=True)
        assert len(front.entities) == 1

        back = m.projected(-m.face_normals[0], ignore_sign=False)
        assert len(back.entities) == 0
        back = m.projected(-m.face_normals[0], ignore_sign=True)
        assert len(back.entities) == 1

    def test_project_multi(self):
        mesh = g.trimesh.creation.box() + g.trimesh.creation.box().apply_translation(
            [3, 0, 0]
        )
        proj = mesh.projected(normal=[0, 0, 1])

        assert mesh.body_count == 2
        assert len(proj.root) == 2
        assert g.np.isclose(proj.area, 2.0)

    def test_second_moment(self):
        def rectangle(extents):
            # return the boundary of an origin-centered
            # rectangle as a numpy array
            a = g.np.abs(g.np.array(extents) / 2.0)
            lower, upper = -a, a
            return g.np.array([lower, [upper[0], lower[1]], upper, [lower[0], upper[1]]])

        def poly(bh, bhi=None):
            # return a rectangle centered at the origin
            # as a shapely Polygon
            shell = rectangle(bh)
            if bhi is not None:
                holes = [rectangle(bhi)]
            else:
                holes = []
            return Polygon(shell=shell, holes=holes)

        def poly_corner(bh):
            # return a rectangle with one corner
            # at the origin and the rest in positive space
            shell = rectangle(bh)
            shell += shell.min(axis=0)
            return Polygon(shell=shell)

        def poly_doublecorner(bh):
            # returns two equal sized rectangles as one polygon
            # Same as poly_corner(), but the rectangle gets
            # mirrored by 180 deg at the origin
            # This puts the centroid in the origin with Ixy != 0
            shell_1 = rectangle(bh)
            shell_1 += shell_1.min(axis=0)
            shell_2 = -shell_1
            shell = g.np.concatenate(
                (shell_1[2:, :], shell_1[:2, :], shell_2[2:, :], shell_2[:2, :]), axis=0
            )
            return Polygon(shell=shell)

        def truth(bh, bhi=None):
            # return the analytical second moment of area
            # for a rectangle with centroid at the origin
            # and width-height of `bh` and an interior
            # rectangle width-height of `bhi`
            if bhi is None:
                bhi = g.np.zeros(2)
            b, h = bh
            bi, hi = bhi
            return (
                g.np.array(
                    [b * h**3 - bi * hi**3, h * b**3 - hi * bi**3, 0.0],
                    dtype=g.np.float64,
                )
                / 12
            )

        def truth_corner(bh):
            # check a rectangle with one corner
            # at the origin and the rest in positive space
            b, h = bh
            return g.np.array(
                [b * h**3 / 3.0, h * b**3 / 3.0, 0.5 * b**2 * 0.5 * h**2],
                dtype=g.np.float64,
            )

        from shapely.geometry import Polygon

        from trimesh.path.polygons import second_moments, transform_polygon

        heights = g.np.array([[0.01, 0.01], [1, 1], [10, 2], [3, 21]])
        for bh in heights:
            # check the second moment of a rectangle
            # as polygon is already centered, centered doesn't have any effect
            O_moments, O_principal_moments, O_alpha, O_transform = second_moments(
                poly(bh), return_centered=True
            )
            # check against wikipedia
            t = truth(bh)
            # for a centered rectangle, the principal axis are already aligned
            # with the frame axis
            assert g.np.allclose(O_moments, t)
            assert g.np.any(g.np.isclose(O_moments, O_principal_moments[0]))
            assert g.np.isclose(O_moments[2], 0)  # Ixy = 0
            assert g.np.isclose(O_alpha, 0)
            assert g.np.allclose(O_transform, g.np.eye(3))

            # now check a rectangle with the corner, so Ixy != 0

            # First we test with centering. The results should be same as
            # with the initially centered rectangles
            C_moments, C_principal_moments, C_alpha, C_transform = second_moments(
                poly_corner(bh), return_centered=True
            )
            assert g.np.allclose(O_moments, C_moments)
            assert g.np.allclose(O_principal_moments, C_principal_moments)
            assert g.np.isclose(O_alpha, C_alpha)
            assert g.np.allclose(O_transform[:, :2], C_transform[:, :2])

            # Now without centering
            moments = second_moments(poly_corner(bh), return_centered=False)
            t = truth_corner(bh)
            assert g.np.allclose(moments, t)

            # Now we will get the transform for a double rectangle. Then we will apply
            # the transform and test if Ixy == 0, alpha == 0 etc.
            C_moments, C_principal_moments, C_alpha, C_transform = second_moments(
                poly_doublecorner(bh), return_centered=True
            )
            # apply the outputted transform to the polygon
            T_polygon = transform_polygon(poly_doublecorner(bh), C_transform)
            # call the function on the transformed polygon
            T_moments, T_principal_moments, T_alpha, T_transform = second_moments(
                T_polygon, return_centered=True
            )
            assert g.np.any(g.np.isclose(T_moments, C_principal_moments[0]))
            assert g.np.allclose(C_principal_moments, T_principal_moments)
            assert g.np.isclose(T_alpha, 0, atol=1e-7)
            assert g.np.allclose(T_transform, g.np.eye(3))

            # check polygons with interior
            for bhi in heights:
                # only check if interior is smaller than exterior
                if not (bhi < bh).all():
                    continue

                # check a rectangle with interiors
                c = second_moments(poly(bh, bhi), return_centered=False)
                t = truth(bh, bhi)
                assert g.np.allclose(c, t)

    def test_native_centroid(self):
        # check our native implementation of the shoelace algorithm
        # against the results from shapely which we use for native
        # checks of counter-clockwise 2D coordinate loops

        # get some polygons without interiors
        polygons = [g.Polygon(i) for i in g.data["nestable"]]
        # same data as (n, 2) float arrays
        coords = [g.np.array(i) for i in g.data["nestable"]]

        for p, c in zip(polygons, coords):
            # area will be signed with respect to counter-clockwise
            _ccw, area, centroid = g.trimesh.util.is_ccw(c, return_all=True)
            assert g.np.allclose(centroid, g.np.array(p.centroid.coords)[0])
            assert g.np.isclose(abs(area), p.area)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
lighting.py
--------------

Hold basic information about lights.

Forked from the light model in `pyrender`:
https://github.com/mmatl/pyrender
"""

import numpy as np

from .. import transformations, util, visual

# default light color
_DEFAULT_RGBA = np.array([60, 60, 60, 255], dtype=np.uint8)


class Light(util.ABC):
    """
    Base class for all light objects.

    Attributes
    ----------
    name : str, optional
        Name of the light.
    color : (4,) uint8
        RGBA value for the light's color in linear space.
    intensity : float
        Brightness of light. The units that this is defined in depend
        on the type of light: point and spot lights use luminous intensity
        in candela (lm/sr) while directional lights use illuminance
        in lux (lm/m2).
    radius : float
        Cutoff distance at which light's intensity may be considered to
        have reached zero. Supported only for point and spot lights
        Must be > 0.0
        If None, the radius is assumed to be infinite.
    """

    def __init__(self, name=None, color=None, intensity=None, radius=None):
        if name is None:
            # if name is not passed, make it something unique
            self.name = f"light_{util.unique_id(6).upper()}"
        else:
            # otherwise assign it
            self.name = name

        self.color = color
        self.intensity = intensity
        self.radius = radius

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        if value is None:
            self._color = _DEFAULT_RGBA
        else:
            value = visual.to_rgba(value)
            if len(value.shape) == 2:
                value = value[0]
            if value.shape != (4,):
                raise ValueError("couldn't convert color to RGBA!")
            # uint8 RGB color
            self._color = value

    @property
    def intensity(self):
        return self._intensity

    @intensity.setter
    def intensity(self, value):
        if value is not None:
            self._intensity = float(value)
        else:
            self._intensity = 1.0

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, value):
        if value is None or value < 0.0:
            self._radius = value
        else:
            self._radius = float(value)


class DirectionalLight(Light):
    """
    Directional lights are light sources that act as though they are
    infinitely far away and emit light in the direction of the local -z axis.
    This light type inherits the orientation of the node that it belongs to;
    position and scale are ignored except for their effect on the inherited
    node orientation. Because it is at an infinite distance, the light is
    not attenuated. Its intensity is defined in lumens per metre squared,
    or lux (lm/m2).

    Attributes
    ----------
    name : str, optional
        Name of the light.
    color : (4,) unit8
        RGBA value for the light's color in linear space.
    intensity : float
        Brightness of light. The units that this is defined in depend
        on the type of light.
        point and spot lights use luminous intensity in candela (lm/sr),
        while directional lights use illuminance in lux (lm/m2).
    radius : float
        Cutoff distance at which light's intensity may be considered to
        have reached zero. Supported only for point and spot lights, must be > 0.
        If None, the radius is assumed to be infinite.
    """

    def __init__(self, name=None, color=None, intensity=None, radius=None):
        super().__init__(name=name, color=color, intensity=intensity, radius=radius)


class PointLight(Light):
    """
    Point lights emit light in all directions from their position in space;
    rotation and scale are ignored except for their effect on the inherited
    node position. The brightness of the light attenuates in a physically
    correct manner as distance increases from the light's position (i.e.
    brightness goes like the inverse square of the distance). Point light
    intensity is defined in candela, which is lumens per square radian (lm/sr).

    Attributes
    ----------
    name : str, optional
        Name of the light.
    color : (4,) uint8
        RGBA value for the light's color in linear space.
    intensity : float
        Brightness of light. The units that this is defined in depend
        on the type of light.
        point and spot lights use luminous intensity in candela (lm/sr),
        while directional lights use illuminance in lux (lm/m2).
    radius : float
        Cutoff distance at which light's intensity may be considered to
        have reached zero. Supported only for point and spot lights, must be > 0.
        If None, the radius is assumed to be infinite.
    """

    def __init__(self, name=None, color=None, intensity=None, radius=None):
        super().__init__(name=name, color=color, intensity=intensity, radius=radius)


class SpotLight(Light):
    """
    Spot lights emit light in a cone in the direction of the local -z axis.
    The angle and falloff of the cone is defined using two numbers, the
    `innerConeAngle` and `outerConeAngle`. As with point lights, the brightness
    also attenuates in a physically correct manner as distance increases from
    the light's position (i.e. brightness goes like the inverse square of the
    distance). Spot light intensity refers to the brightness inside the
    `innerConeAngle` (and at the location of the light) and is defined in
    candela, which is lumens per square radian (lm/sr). A spot light's position
    and orientation are inherited from its node transform. Inherited scale does
    not affect cone shape, and is ignored except for its effect on position
    and orientation.

    Attributes
    ----------
    name : str, optional
        Name of the light.
    color : (4,) uint8
        RGBA value for the light's color in linear space.
    intensity : float
        Brightness of light. The units that this is defined in depend
        on the type of light.
        point and spot lights use luminous intensity in candela (lm/sr),
        while directional lights use illuminance in lux (lm/m2).
    radius : float
        Cutoff distance at which light's intensity may be considered to
        have reached zero. Supported only for point and spot lights, must be > 0.
        If None, the radius is assumed to be infinite.
    innerConeAngle : float
        Angle, in radians, from centre of spotlight where falloff begins.
        Must be greater than or equal to `0` and less than `outerConeAngle`.
    outerConeAngle : float
        Angle, in radians, from centre of spotlight where falloff ends.
        Must be greater than `innerConeAngle` and less than or equal to `PI / 2.0`.
    """

    def __init__(
        self,
        name=None,
        color=None,
        intensity=None,
        radius=None,
        innerConeAngle=0.0,
        outerConeAngle=np.pi / 4.0,
    ):
        super().__init__(name=name, color=color, intensity=intensity, radius=radius)
        self.outerConeAngle = outerConeAngle
        self.innerConeAngle = innerConeAngle

    @property
    def innerConeAngle(self):
        return self._innerConeAngle

    @innerConeAngle.setter
    def innerConeAngle(self, value):
        if value < 0.0 or value > self.outerConeAngle:
            raise ValueError("Invalid value for inner cone angle")
        self._innerConeAngle = float(value)

    @property
    def outerConeAngle(self):
        return self._outerConeAngle

    @outerConeAngle.setter
    def outerConeAngle(self, value):
        if value < 0.0 or value > np.pi / 2.0 + 1e-9:
            raise ValueError("Invalid value for outer cone angle")
        self._outerConeAngle = float(value)


def autolight(scene):
    """
    Generate a list of lights for a scene that looks decent.

    Parameters
    --------------
    scene : trimesh.Scene
      Scene with geometry

    Returns
    --------------
    lights : [Light]
      List of light objects
    transforms : (len(lights), 4, 4) float
      Transformation matrices for light positions.
    """

    # create two default point lights
    lights = [PointLight(), PointLight()]

    # create two translation matrices for bounds corners
    transforms = [transformations.translation_matrix(b) for b in scene.bounds]

    return lights, transforms



from ... import util
from ...exceptions import ExceptionWrapper
from ...exchange.ply import load_ply
from ...typed import Optional, Set
from ..path import Path
from . import misc
from .dxf import _dxf_loaders
from .svg_io import _svg_loaders


def load_path(file_obj, file_type: Optional[str] = None, **kwargs):
    """
    Load a file to a Path file_object.

    Parameters
    -----------
    file_obj
      Accepts many types:
         - Path, Path2D, or Path3D file_objects
         - open file file_object (dxf or svg)
         - file name (dxf or svg)
         - shapely.geometry.Polygon
         - shapely.geometry.MultiLineString
         - dict with kwargs for Path constructor
         - `(n, 2, (2|3)) float` line segments
    file_type
        Type of file is required if file
        object is passed.

    Returns
    ---------
    path : Path, Path2D, Path3D file_object
      Data as a native trimesh Path file_object
    """
    # avoid a circular import
    from ...exchange.load import _load_kwargs, _parse_file_args

    arg = _parse_file_args(file_obj=file_obj, file_type=file_type, **kwargs)

    if isinstance(file_obj, Path):
        # we have been passed a file object that is already a loaded
        # trimesh.path.Path object so do nothing and return
        return file_obj
    elif util.is_file(arg.file_obj):
        if arg.file_type in path_loaders:
            kwargs.update(
                path_loaders[arg.file_type](
                    file_obj=arg.file_obj, file_type=arg.file_type
                )
            )
        elif arg.file_type == "ply":
            # we cannot register this exporter to path_loaders since
            # this is already reserved by Trimesh in ply format in trimesh.load()
            kwargs.update(load_ply(file_obj=arg.file_obj, file_type=arg.file_type))
    elif util.is_instance_named(file_obj, ["Polygon", "MultiPolygon"]):
        # convert from shapely polygons to Path2D
        kwargs.update(misc.polygon_to_path(file_obj))
    elif util.is_instance_named(file_obj, "MultiLineString"):
        # convert from shapely LineStrings to Path2D
        kwargs.update(misc.linestrings_to_path(file_obj))
    elif isinstance(file_obj, dict):
        # load as kwargs
        kwargs = file_obj
    elif util.is_sequence(file_obj):
        # load as lines in space
        kwargs.update(misc.lines_to_path(file_obj))
    else:
        raise ValueError("Not a supported object type!")

    # actually load
    result = _load_kwargs(kwargs)
    result._source = arg

    return result


def path_formats() -> Set[str]:
    """
    Get a list of supported path formats.

    Returns
    ------------
    loaders
       Extensions of loadable formats, i.e. {'svg', 'dxf'}
    """

    return {k for k, v in path_loaders.items() if not isinstance(v, ExceptionWrapper)}


path_loaders = {}
path_loaders.update(_svg_loaders)
path_loaders.update(_dxf_loaders)



"""
caching.py
-----------

Functions and classes that help with tracking changes
in `numpy.ndarray` and clearing cached values based
on those changes.

You should really `pip install xxhash`:

```
In [23]: %timeit int(blake2b(d).hexdigest(), 16)
102 us +/- 684 ns per loop

In [24]: %timeit int(sha256(d).hexdigest(), 16)
142 us +/- 3.73 us

In [25]: %timeit xxh3_64_intdigest(d)
3.37 us +/- 116 ns per loop
```
"""

import os
import sys
import time
from functools import wraps
from hashlib import sha256 as _sha256

import numpy as np

from .constants import log
from .util import is_sequence

try:
    from collections.abc import Mapping
except BaseException:
    from collections.abc import Mapping


def sha256(item) -> int:
    return int(_sha256(item).hexdigest(), 16)


if sys.version_info >= (3, 9):
    # blake2b is available on Python 3 and
    from hashlib import blake2b as _blake2b

    def hash_fallback(item):
        return int(_blake2b(item, usedforsecurity=False).hexdigest(), 16)
else:
    # fallback to sha256
    hash_fallback = sha256

# xxhash is up to 30x faster than sha256:
# `pip install xxhash`
try:
    # newest version of algorithm
    from xxhash import xxh3_64_intdigest as hash_fast
except BaseException:
    try:
        # older version of the algorithm
        from xxhash import xxh64_intdigest as hash_fast
    except BaseException:
        # use hashlib as a fallback hashing library
        log.debug(
            "falling back to hashlib "
            + "hashing: `pip install xxhash`"
            + "for 50x faster cache checks"
        )
        hash_fast = hash_fallback


def tracked_array(array, dtype=None):
    """
    Properly subclass a numpy ndarray to track changes.

    Avoids some pitfalls of subclassing by forcing contiguous
    arrays and does a view into a TrackedArray.

    Parameters
    ------------
    array : array- like object
      To be turned into a TrackedArray
    dtype : np.dtype
      Which dtype to use for the array

    Returns
    ------------
    tracked : TrackedArray
      Contains input array data.
    """
    # if someone passed us None, just create an empty array
    if array is None:
        array = []
    # make sure it is contiguous then view it as our subclass
    tracked = np.ascontiguousarray(array, dtype=dtype).view(TrackedArray)
    # should always be contiguous here
    assert tracked.flags["C_CONTIGUOUS"]

    return tracked


def cache_decorator(function):
    """
    A decorator for class methods, replaces @property
    but will store and retrieve function return values
    in object cache.

    Parameters
    ------------
    function : method
      This is used as a decorator:
      ```
      @cache_decorator
      def foo(self, things):
        return 'happy days'
      ```
    """

    # use wraps to preserve docstring
    @wraps(function)
    def get_cached(*args, **kwargs):
        """
        Only execute the function if its value isn't stored
        in cache already.
        """
        self = args[0]
        # use function name as key in cache
        name = function.__name__
        # do the dump logic ourselves to avoid
        # verifying cache twice per call
        self._cache.verify()
        # access cache dict to avoid automatic validation
        # since we already called cache.verify manually
        if name in self._cache.cache:
            # already stored so return value
            return self._cache.cache[name]
        # value not in cache so execute the function
        value = function(*args, **kwargs)
        # store the value
        if (
            self._cache.force_immutable
            and hasattr(value, "flags")
            and len(value.shape) > 0
        ):
            value.flags.writeable = False

        self._cache.cache[name] = value

        return value

    # all cached values are also properties
    # so they can be accessed like value attributes
    # rather than functions
    return property(get_cached)


class TrackedArray(np.ndarray):
    """
    Subclass of numpy.ndarray that provides hash methods
    to track changes.

    General method is to aggressively set 'modified' flags
    on operations which might (but don't necessarily) alter
    the array, ideally we sometimes compute hashes when we
    don't need to, but we don't return wrong hashes ever.

    We store boolean modified flag for each hash type to
    make checks fast even for queries of different hashes.

    Methods
    ----------
    __hash__ : int
      Runs the fastest available hash in this order:
        `xxh3_64, xxh_64, blake2b, sha256`
    """

    def __array_finalize__(self, obj):
        """
        Sets a modified flag on every TrackedArray
        This flag will be set on every change as well as
        during copies and certain types of slicing.
        """

        self._dirty_hash = True
        if isinstance(obj, type(self)):
            obj._dirty_hash = True

    def __array_wrap__(self, out_arr, context=None, *args, **kwargs):
        """
        Return a numpy scalar if array is 0d.
        See https://github.com/numpy/numpy/issues/5819
        """
        if out_arr.ndim:
            return np.ndarray.__array_wrap__(self, out_arr, context, *args, **kwargs)
        # Match numpy's behavior and return a numpy dtype scalar
        return out_arr[()]

    @property
    def mutable(self):
        return self.flags["WRITEABLE"]

    @mutable.setter
    def mutable(self, value):
        self.flags.writeable = value

    def __hash__(self):
        """
        Return a fast hash of the contents of the array.

        Returns
        -------------
        hash : long int
          A hash of the array contents.
        """
        # repeat the bookkeeping to get a contiguous array
        if not self._dirty_hash and hasattr(self, "_hashed"):
            # we have a valid hash without recomputing.
            return self._hashed

        # run a hashing function on the C-order bytes copy
        hashed = hash_fast(self.tobytes(order="C"))

        # assign the value and set the flag
        self._hashed = hashed
        self._dirty_hash = False

        return hashed

    def __iadd__(self, *args, **kwargs):
        """
        In-place addition.

        The i* operations are in- place and modify the array,
        so we better catch all of them.
        """

        self._dirty_hash = True
        return super(self.__class__, self).__iadd__(*args, **kwargs)

    def __isub__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__isub__(*args, **kwargs)

    def fill(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).fill(*args, **kwargs)

    def partition(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).partition(*args, **kwargs)

    def put(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).put(*args, **kwargs)

    def byteswap(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).byteswap(*args, **kwargs)

    def itemset(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).itemset(*args, **kwargs)

    def sort(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).sort(*args, **kwargs)

    def setflags(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).setflags(*args, **kwargs)

    def __imul__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__imul__(*args, **kwargs)

    def __idiv__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__idiv__(*args, **kwargs)

    def __itruediv__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__itruediv__(*args, **kwargs)

    def __imatmul__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__imatmul__(*args, **kwargs)

    def __ipow__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__ipow__(*args, **kwargs)

    def __imod__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__imod__(*args, **kwargs)

    def __ifloordiv__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__ifloordiv__(*args, **kwargs)

    def __ilshift__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__ilshift__(*args, **kwargs)

    def __irshift__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__irshift__(*args, **kwargs)

    def __iand__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__iand__(*args, **kwargs)

    def __ixor__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__ixor__(*args, **kwargs)

    def __ior__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__ior__(*args, **kwargs)

    def __setitem__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__setitem__(*args, **kwargs)

    def __setslice__(self, *args, **kwargs):
        self._dirty_hash = True
        return super(self.__class__, self).__setslice__(*args, **kwargs)


class Cache:
    """
    Class to cache values which will be stored until the
    result of an ID function changes.
    """

    def __init__(self, id_function, force_immutable=False):
        """
        Create a cache object.

        Parameters
        ------------
        id_function : function
          Returns hashable value
        force_immutable : bool
          If set will make all numpy arrays read-only
        """
        self._id_function = id_function
        # for stored numpy arrays set `flags.writable = False`
        self.force_immutable = bool(force_immutable)
        # call the id function for initial value
        self.id_current = None
        # a counter for locks
        self._lock = 0
        # actual store for data
        self.cache = {}

    def delete(self, key):
        """
        Remove a key from the cache.
        """
        if key in self.cache:
            self.cache.pop(key, None)

    def verify(self):
        """
        Verify that the cached values are still for the same
        value of id_function and delete all stored items if
        the value of id_function has changed.
        """
        # if we are in a lock don't check anything
        if self._lock != 0:
            return

        # check the hash of our data
        id_new = self._id_function()

        # things changed
        if id_new != self.id_current:
            if len(self.cache) > 0:
                log.debug(
                    "%d items cleared from cache: %s",
                    len(self.cache),
                    str(list(self.cache.keys())),
                )
            # hash changed, so dump the cache
            # do it manually rather than calling clear()
            # as we are internal logic and can avoid function calls
            self.cache = {}
            # set the id to the new data hash
            self.id_current = id_new

    def clear(self, exclude=None):
        """
        Remove elements in the cache.

        Parameters
        -----------
        exclude : list
          List of keys in cache to not clear.
        """
        if exclude is None:
            self.cache = {}
        else:
            self.cache = {k: v for k, v in self.cache.items() if k in exclude}

    def update(self, items):
        """
        Update the cache with a set of key, value pairs without
        checking id_function.
        """
        self.cache.update(items)

        if self.force_immutable:
            for v in self.cache.values():
                if hasattr(v, "flags") and len(v.shape) > 0:
                    v.flags.writeable = False
        self.id_set()

    def id_set(self):
        """
        Set the current ID to the value of the ID function.
        """
        self.id_current = self._id_function()

    def __getitem__(self, key):
        """
        Get an item from the cache. If the item
        is not in the cache, it will return None

        Parameters
        -------------
        key : hashable
               Key in dict

        Returns
        -------------
        cached : object, or None
          Object that was stored
        """
        self.verify()
        if key in self.cache:
            return self.cache[key]
        return None

    def __setitem__(self, key, value):
        """
        Add an item to the cache.

        Parameters
        ------------
        key : hashable
          Key to reference value
        value : any
          Value to store in cache
        """
        # dumpy cache if ID function has changed
        self.verify()
        # make numpy arrays read-only if asked to
        if self.force_immutable and hasattr(value, "flags") and len(value.shape) > 0:
            value.flags.writeable = False
        # assign data to dict
        self.cache[key] = value

        return value

    def __contains__(self, key):
        self.verify()
        return key in self.cache

    def __len__(self):
        self.verify()
        return len(self.cache)

    def __enter__(self):
        self._lock += 1

    def __exit__(self, *args):
        self._lock -= 1
        self.id_current = self._id_function()


class DiskCache:
    """
    Store results of expensive operations on disk
    with an option to expire the results. This is used
    to cache the multi-gigabyte test corpuses in
    `tests/corpus.py`
    """

    def __init__(self, path, expire_days=30):
        """
        Create a cache on disk for storing expensive results.

        Parameters
        --------------
        path : str
          A writeable location on the current file path.
        expire_days : int or float
          How old should results be considered expired.

        """
        # store how old we allow results to be
        self.expire_days = expire_days
        # store the location for saving results
        self.path = os.path.abspath(os.path.expanduser(path))
        # make sure the specified path exists
        os.makedirs(self.path, exist_ok=True)

    def get(self, key, fetch):
        """
        Get a key from the cache or run a calculation.

        Parameters
        -----------
        key : str
          Key to reference item with
        fetch : function
          If key isn't stored and recent run this
          function and store its result on disk.
        """
        # hash the key so we have a fixed length string
        key_hash = _sha256(key.encode("utf-8")).hexdigest()
        # full path of result on local disk
        path = os.path.join(self.path, key_hash)

        # check to see if we can use the cache
        if os.path.isfile(path):
            # compute the age of the existing file in days
            age_days = (time.time() - os.stat(path).st_mtime) / 86400.0
            if age_days < self.expire_days:
                # this nested condition means that
                # the file both exists and is recent
                # enough, so just return its contents
                with open(path, "rb") as f:
                    return f.read()

        log.debug(f"not in cache fetching: `{key}`")
        # since we made it here our data isn't cached
        # run the expensive function to fetch the file
        raw = fetch()
        # write the data so we can save it
        with open(path, "wb") as f:
            f.write(raw)

        # return the data
        return raw


class DataStore(Mapping):
    """
    A class to store multiple numpy arrays and track them all
    for changes.

    Operates like a dict that only stores numpy.ndarray
    """

    def __init__(self):
        self.data = {}

    def __iter__(self):
        return iter(self.data)

    def pop(self, key):
        return self.data.pop(key, None)

    def __delitem__(self, key):
        self.data.pop(key, None)

    @property
    def mutable(self):
        """
        Is data allowed to be altered or not.

        Returns
        -----------
        is_mutable : bool
          Can data be altered in the DataStore
        """
        return getattr(self, "_mutable", True)

    @mutable.setter
    def mutable(self, value):
        """
        Is data allowed to be altered or not.

        Parameters
        ------------
        is_mutable : bool
          Should data be allowed to be altered
        """
        # make sure passed value is a bool
        is_mutable = bool(value)
        # apply the flag to any data stored
        for v in self.data.values():
            if isinstance(v, TrackedArray):
                v.mutable = value
        # save the mutable setting
        self._mutable = is_mutable

    def is_empty(self):
        """
        Is the current DataStore empty or not.

        Returns
        ----------
        empty : bool
          False if there are items in the DataStore
        """
        if len(self.data) == 0:
            return True
        for v in self.data.values():
            if is_sequence(v):
                if len(v) == 0:
                    return True
                else:
                    return False
            elif bool(np.isreal(v)):
                return False
        return True

    def clear(self):
        """
        Remove all data from the DataStore.
        """
        self.data = {}

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, data):
        """
        Store an item in the DataStore.

        Parameters
        -------------
        key
          A hashable key to store under
        data
          Usually a numpy array which will be subclassed
          but anything hashable should be able to be stored.
        """
        # we shouldn't allow setting on immutable datastores
        if not self.mutable:
            raise ValueError("DataStore is configured immutable!")

        if isinstance(data, TrackedArray):
            # don't bother to re-track TrackedArray
            tracked = data
        elif isinstance(data, (np.ndarray, list, set, tuple)):
            # wrap data if it is array-like
            tracked = tracked_array(data)
        else:
            try:
                # will raise if this is not a hashable type
                hash(data)
            except BaseException:
                raise ValueError(f"unhashable `{key}:{type(data)}`")
            tracked = data

        # apply our mutability setting
        if hasattr(self, "_mutable"):
            # apply our mutability setting only if it was explicitly set
            tracked.mutable = self.mutable
        # store data
        self.data[key] = tracked

    def __contains__(self, key):
        return key in self.data

    def __len__(self):
        return len(self.data)

    def update(self, values):
        if not isinstance(values, dict):
            raise ValueError("Update only implemented for dicts")
        for key, value in values.items():
            self[key] = value

    def __hash__(self):
        """
        Get a hash reflecting everything in the DataStore.

        Returns
        ----------
        hash : str
          hash of data in hexadecimal
        """
        # only hash values that aren't None
        # or if they are arrays require length greater than zero
        return hash_fast(
            np.array(
                [
                    hash(v)
                    for v in self.data.values()
                    if v is not None and (not hasattr(v, "__len__") or len(v) > 0)
                ],
                dtype=np.int64,
            ).tobytes()
        )



import numpy as np

from .. import util


class HeaderError(Exception):
    # the exception raised if an STL file object doesn't match its header
    pass


# define a numpy datatype for the data section of a binary STL file
# everything in STL is always Little Endian
# this works natively on Little Endian systems, but blows up on Big Endians
# so we always specify byteorder
_stl_dtype = np.dtype(
    [("normals", "<f4", (3)), ("vertices", "<f4", (3, 3)), ("attributes", "<u2")]
)
# define a numpy datatype for the header of a binary STL file
_stl_dtype_header = np.dtype([("header", np.void, 80), ("face_count", "<u4")])


def load_stl(file_obj, **kwargs):
    """
    Load an STL file from a file object.

    Parameters
    ----------
    file_obj : open file-like object
      Containing STL data

    Returns
    ----------
    loaded : dict
      kwargs for a Trimesh constructor with keys:
      vertices:     (n,3) float, vertices
      faces:        (m,3) int, indexes of vertices
      face_normals: (m,3) float, normal vector of each face
    """
    # save start of file obj
    file_pos = file_obj.tell()
    try:
        # check the file for a header which matches the file length
        # if that is true, it is almost certainly a binary STL file
        # if the header doesn't match the file length a HeaderError will be
        # raised
        return load_stl_binary(file_obj)
    except HeaderError:
        # move the file back to where it was initially
        file_obj.seek(file_pos)
        # try to load the file as an ASCII STL
        # if the header doesn't match the file length
        # HeaderError will be raised
        return load_stl_ascii(file_obj)


def load_stl_binary(file_obj):
    """
    Load a binary STL file from a file object.

    Parameters
    ----------
    file_obj : open file- like object
      Containing STL data

    Returns
    ----------
    loaded: kwargs for a Trimesh constructor with keys:
              vertices:     (n,3) float, vertices
              faces:        (m,3) int, indexes of vertices
              face_normals: (m,3) float, normal vector of each face
    """
    # the header is always 84 bytes long, we just reference the dtype.itemsize
    # to be explicit about where that magical number comes from
    header_length = _stl_dtype_header.itemsize
    header_data = file_obj.read(header_length)
    if len(header_data) < header_length:
        raise HeaderError("Binary STL shorter than a fixed header!")

    try:
        header = np.frombuffer(header_data, dtype=_stl_dtype_header)
    except BaseException:
        raise HeaderError("Binary header incorrect type")

    try:
        # save the header block as a string
        # there could be any garbage in there so wrap in try
        metadata = {"header": util.decode_text(bytes(header["header"][0])).strip()}
    except BaseException:
        metadata = {}

    # now we check the length from the header versus the length of the file
    # data_start should always be position 84, but hard coding that felt ugly
    data_start = file_obj.tell()
    # this seeks to the end of the file
    # position 0, relative to the end of the file 'whence=2'
    file_obj.seek(0, 2)
    # we save the location of the end of the file and seek back to where we
    # started from
    data_end = file_obj.tell()
    file_obj.seek(data_start)

    # the binary format has a rigidly defined structure, and if the length
    # of the file doesn't match the header, the loaded version is almost
    # certainly going to be garbage.
    len_data = data_end - data_start
    len_expected = header["face_count"] * _stl_dtype.itemsize

    # this check is to see if this really is a binary STL file.
    # if we don't do this and try to load a file that isn't structured properly
    # we will be producing garbage or crashing hard
    # so it's much better to raise an exception here.
    if len_data != len_expected:
        raise HeaderError(
            f"Binary STL has incorrect length in header: {len_data} vs {len_expected}"
        )

    blob = np.frombuffer(file_obj.read(), dtype=_stl_dtype)

    # return empty geometry if there are no vertices
    if not len(blob["vertices"]):
        return {"geometry": {}}

    # all of our vertices will be loaded in order
    # so faces are just sequential indices reshaped.
    faces = np.arange(header["face_count"][0] * 3).reshape((-1, 3))

    # there are two bytes per triangle saved for anything
    # which is sometimes used for face color
    result = {
        "vertices": blob["vertices"].reshape((-1, 3)),
        "face_normals": blob["normals"].reshape((-1, 3)),
        "faces": faces,
        "face_attributes": {"stl": blob["attributes"]},
        "metadata": metadata,
    }
    return result


def load_stl_ascii(file_obj):
    """
    Load an ASCII STL file from a file object.

    Parameters
    ----------
    file_obj : open file- like object
      Containing input data

    Returns
    ----------
    loaded : dict
      kwargs for a Trimesh constructor with keys:
      vertices:     (n, 3) float, vertices
      faces:        (m, 3) int, indexes of vertices
      face_normals: (m, 3) float, normal vector of each face
    """

    # read all text into one string
    raw_mixed = util.decode_text(file_obj.read()).strip()
    # convert to lower case for solids and name capture
    raw_lower = raw_mixed.lower()

    # collect the keyword arguments for the Trimesh constructor
    kwargs = {}

    # keep track of our position in the file
    position = 0

    # use a for loop to avoid any possibility of infinite looping
    for _ in range(len(raw_mixed)):
        # find the start of the solid chunk
        solid_start = raw_lower.find("solid", position)
        # find the end of the solid chunk
        solid_end = raw_lower.find("endsolid", position)

        # on the next loop we don't have to check the text we've consumed
        position = solid_end + len("endsolid")

        # delimiter wasn't found for a chunk so exit
        if solid_end < 0 or solid_start < 0:
            break

        # end delimiter order is wrong so this file is very malformed
        if solid_start > solid_end:
            raise ValueError("`endsolid` precedes `solid`!")

        # get the chunk of text with this particular solid
        solid = raw_lower[solid_start:solid_end]

        # extract the vertices
        vertex_text = solid.split("vertex")
        vertices = np.fromstring(
            " ".join(line[: line.find("\n")] for line in vertex_text[1:]),
            sep=" ",
            dtype=np.float64,
        )
        if len(vertices) < 3:
            continue
        if len(vertices) % 3 != 0:
            raise ValueError("incorrect number of vertices")

        # reshape vertices to final 3D shape
        vertices = vertices.reshape((-1, 3))
        faces = np.arange(len(vertices)).reshape((-1, 3))

        # try to extract the face normals the same way
        face_normals = None
        try:
            normal_text = solid.split("normal")
            normals = np.fromstring(
                " ".join(line[: line.find("\n")] for line in normal_text[1:]),
                sep=" ",
                dtype=np.float64,
            )
            if len(normals) == len(vertices):
                face_normals = normals.reshape((-1, 3))
        except BaseException:
            util.log.warning("failed to extract face_normals", exc_info=True)

        try:
            # Previously checked to make sure there was matching 'solid' for 'endsolid'
            # the name is right after the `solid` keyword if it exists
            name = raw_mixed[solid_start : solid_start + solid.find("\n")][6:].strip()
        except BaseException:
            # will be filled in by unique_name
            name = None

        # make sure geometry has a unique name for the scene
        name = util.unique_name(name, kwargs)
        # save the constructor arguments
        kwargs[name] = {
            "vertices": vertices.reshape((-1, 3)),
            "face_normals": face_normals,
            "faces": faces,
            "metadata": {"name": name},
        }

    if len(kwargs) == 1:
        return next(iter(kwargs.values()))

    return {"geometry": kwargs}


def export_stl(mesh) -> bytes:
    """
    Convert a Trimesh object into a binary STL file.

    Parameters
    ---------
    mesh
      Trimesh object to export.

    Returns
    ---------
    export
      Represents mesh in binary STL form
    """
    header = np.zeros(1, dtype=_stl_dtype_header)
    if hasattr(mesh, "faces"):
        header["face_count"] = len(mesh.faces)
    export = header.tobytes()

    if hasattr(mesh, "faces"):
        packed = np.zeros(len(mesh.faces), dtype=_stl_dtype)
        packed["normals"] = mesh.face_normals
        packed["vertices"] = mesh.triangles
        export += packed.tobytes()

    return export


def export_stl_ascii(mesh) -> str:
    """
    Convert a Trimesh object into an ASCII STL file.

    Parameters
    ---------
    mesh : trimesh.Trimesh

    Returns
    ---------
    export
      Mesh represented as an ASCII STL file
    """

    # move all the data that's going into the STL file into one array
    blob = np.zeros((len(mesh.faces), 4, 3))
    blob[:, 0, :] = mesh.face_normals
    blob[:, 1:, :] = mesh.triangles

    # create a lengthy format string for the data section of the file
    formatter = (
        "\n".join(
            [
                "facet normal {} {} {}",
                "outer loop",
                "vertex {} {} {}\nvertex {} {} {}\nvertex {} {} {}",
                "endloop",
                "endfacet",
                "",
            ]
        )
    ) * len(mesh.faces)

    # try applying the name from metadata if it exists
    name = mesh.metadata.get("name", "")
    if not isinstance(name, str):
        name = ""
    if len(name) > 80 or "\n" in name:
        name = ""

    # concatenate the header, data, and footer, and a new line
    return "\n".join([f"solid {name}", formatter.format(*blob.reshape(-1)), "endsolid\n"])


_stl_loaders = {"stl": load_stl, "stl_ascii": load_stl}



import argparse


def main():
    """
    A simple command line utility for accessing trimesh functions.

    To display a mesh:
      > trimesh hi.stl

    To convert a mesh:
      > trimesh hi.stl -e hey.glb

    To print some information about a mesh:
      > trimesh hi.stl --statistics
    """
    from .exchange.load import load

    parser = argparse.ArgumentParser()
    parser.add_argument("file_name", nargs="?")

    parser.add_argument(
        "-i",
        "--interact",
        action="store_true",
        help="Get an interactive terminal with trimesh and loaded geometry",
    )
    parser.add_argument("-e", "--export", help="Export a loaded geometry to a new file.")

    args = parser.parse_args()

    if args.file_name is None:
        parser.print_help()
        return
    else:
        scene = load(args.file_name)

    summary(scene)

    if args.export is not None:
        scene.export(args.export)

    if args.interact:
        return interactive(scene)

    scene.show()


def summary(geom):
    """ """
    print(geom)


def interactive(scene):
    """
    Run an interactive session with a loaded scene and trimesh.

    This uses the standard library `code.InteractiveConsole`
    """
    local = locals()

    from code import InteractiveConsole

    # filter out junk variables.
    InteractiveConsole(locals={k: v for k, v in local.items() if k != "local"}).interact()


if __name__ == "__main__":
    main()



import collections
import subprocess
import tempfile
from string import Template

import numpy as np
from numpy.lib.recfunctions import structured_to_unstructured, unstructured_to_structured

from .. import grouping, resources, util, visual
from ..constants import log
from ..geometry import triangulate_quads
from ..resolvers import Resolver
from ..typed import NDArray, Optional

# from ply specification, and additional dtypes found in the wild
_dtypes = {
    "char": "i1",
    "uchar": "u1",
    "short": "i2",
    "ushort": "u2",
    "int": "i4",
    "int8": "i1",
    "int16": "i2",
    "int32": "i4",
    "int64": "i8",
    "uint": "u4",
    "uint8": "u1",
    "uint16": "u2",
    "uint32": "u4",
    "uint64": "u8",
    "float": "f4",
    "float16": "f2",
    "float32": "f4",
    "float64": "f8",
    "double": "f8",
}

# Inverse of the above dict, collisions on numpy type were removed
_inverse_dtypes = {
    "i1": "char",
    "u1": "uchar",
    "i2": "short",
    "u2": "ushort",
    "i4": "int",
    "i8": "int64",
    "u4": "uint",
    "u8": "uint64",
    "f4": "float",
    "f2": "float16",
    "f8": "double",
}


def _numpy_type_to_ply_type(_numpy_type):
    """
    Returns the closest ply equivalent of a numpy type

    Parameters
    ---------
    _numpy_type : a numpy datatype

    Returns
    ---------
    ply_type : string
    """
    return _inverse_dtypes[_numpy_type.str[1:]]


def load_ply(
    file_obj,
    resolver: Optional[Resolver] = None,
    fix_texture: bool = True,
    prefer_color: Optional[str] = None,
    skip_materials: bool = False,
    *args,
    **kwargs,
):
    """
    Load a PLY file from an open file object.

    Parameters
    ---------
    file_obj : an open file- like object
      Source data, ASCII or binary PLY
    resolver
      Object which can resolve assets
    fix_texture
      If True, will re- index vertices and faces
      so vertices with different UV coordinates
      are disconnected.
    skip_materials
      If True, will not load texture (if present).
    prefer_color
      None, 'vertex', or 'face'
      Which kind of color to prefer if both defined

    Returns
    ---------
    mesh_kwargs : dict
      Data which can be passed to
      Trimesh constructor, eg: a = Trimesh(**mesh_kwargs)
    """

    # OrderedDict which is populated from the header
    elements, is_ascii, image_name = _parse_header(file_obj)

    # functions will fill in elements from file_obj
    if is_ascii:
        _ply_ascii(elements, file_obj)
    else:
        _ply_binary(elements, file_obj)

    # try to load the referenced image
    image = None
    if not skip_materials:
        try:
            # soft dependency
            import PIL.Image

            # if an image name is passed try to load it
            if image_name is not None:
                data = resolver.get(image_name)
                image = PIL.Image.open(util.wrap_as_stream(data))
        except ImportError:
            log.debug("textures require `pip install pillow`")
        except BaseException:
            log.warning("unable to load image!", exc_info=True)

    # translate loaded PLY elements to kwargs
    kwargs = _elements_to_kwargs(
        image=image, elements=elements, fix_texture=fix_texture, prefer_color=prefer_color
    )

    return kwargs


def _add_attributes_to_dtype(dtype, attributes):
    """
    Parses attribute datatype to populate a numpy dtype list

    Parameters
    ----------
    dtype : list of numpy datatypes
      operated on in place
    attributes : dict
      contains all the attributes to parse

    Returns
    ----------
    dtype : list of numpy datatypes
    """
    for name, data in attributes.items():
        if data.ndim == 1:
            dtype.append((name, data.dtype))
        else:
            attribute_dtype = data.dtype if len(data.dtype) == 0 else data.dtype[0]
            dtype.append((f"{name}_count", "<u1"))
            dtype.append((name, attribute_dtype, data.shape[1]))
    return dtype


def _add_attributes_to_header(header, attributes):
    """
    Parses attributes in to ply header entries

    Parameters
    ----------
    header : list of ply header entries
      operated on in place
    attributes : dict
      contains all the attributes to parse

    Returns
    ----------
    header : list
      Contains ply header entries
    """
    for name, data in attributes.items():
        if data.ndim == 1:
            header.append(f"property {_numpy_type_to_ply_type(data.dtype)} {name}\n")
        else:
            header.append(
                f"property list uchar {_numpy_type_to_ply_type(data.dtype)} {name}\n"
            )
    return header


def _add_attributes_to_data_array(data_array, attributes):
    """
    Parses attribute data in to a custom array, assumes datatype has been defined
    appropriately

    Parameters
    ----------
    data_array : numpy array with custom datatype
      datatype reflects all the data to be stored for a given ply element
    attributes : dict
      contains all the attributes to parse

    Returns
    ----------
    data_array : numpy array with custom datatype
    """
    for name, data in attributes.items():
        if data.ndim > 1:
            data_array[f"{name}_count"] = data.shape[1] * np.ones(data.shape[0])
        data_array[name] = data
    return data_array


def _assert_attributes_valid(attributes):
    """
    Asserts that a set of attributes is valid for PLY export.

    Parameters
    ----------
    attributes : dict
      Contains the attributes to validate

    Raises
    --------
    ValueError
      If passed attributes aren't valid.
    """
    for data in attributes.values():
        if data.ndim not in [1, 2]:
            raise ValueError("PLY attributes are limited to 1 or 2 dimensions")
        # Inelegant test for structured arrays, reference:
        # https://numpy.org/doc/stable/user/basics.rec.html
        if data.dtype.names is not None:
            raise ValueError("PLY attributes must be of a single datatype")


def export_ply(
    mesh,
    encoding="binary",
    vertex_normal: Optional[bool] = None,
    include_attributes: bool = True,
):
    """
    Export a mesh in the PLY format.

    Parameters
    ----------
    mesh : trimesh.Trimesh
      Mesh to export.
    encoding : str
      PLY encoding: 'ascii' or 'binary_little_endian'
    vertex_normal : None or include vertex normals

    Returns
    ----------
    export : bytes of result
    """
    # evaluate input args
    # allow a shortcut for binary
    if encoding == "binary":
        encoding = "binary_little_endian"
    elif encoding not in ["binary_little_endian", "ascii"]:
        raise ValueError("encoding must be binary or ascii")
    # if vertex normals aren't specifically asked for
    # only export them if they are stored in cache
    if vertex_normal is None:
        vertex_normal = "vertex_normals" in mesh._cache

    # if we want to include mesh attributes in the export
    if include_attributes:
        if hasattr(mesh, "vertex_attributes"):
            # make sure to export texture coordinates as well
            if (
                hasattr(mesh, "visual")
                and hasattr(mesh.visual, "uv")
                and np.shape(mesh.visual.uv) == (len(mesh.vertices), 2)
            ):
                mesh.vertex_attributes["s"] = mesh.visual.uv[:, 0]
                mesh.vertex_attributes["t"] = mesh.visual.uv[:, 1]
            _assert_attributes_valid(mesh.vertex_attributes)
        if hasattr(mesh, "face_attributes"):
            _assert_attributes_valid(mesh.face_attributes)

    # custom numpy dtypes for exporting
    dtype_face = [("count", "<u1"), ("index", "<i4", (3))]
    dtype_vertex = [("vertex", "<f4", (3))]
    # will be appended to main dtype if needed
    dtype_vertex_normal = ("normals", "<f4", (3))
    dtype_color = ("rgba", "<u1", (4))
    # for Path objects.
    dtype_edge = [("index", "<i4", (2))]

    # get template strings in dict
    templates = resources.get_json("templates/ply.json")
    # start collecting elements into a string for the header
    header = [templates["intro"]]
    header_params = {"encoding": encoding}

    # structured arrays for exports
    pack_edges: Optional[NDArray] = None
    pack_vertex: Optional[NDArray] = None
    pack_faces: Optional[NDArray] = None

    # check if scene has geometry
    # check if this is a `trimesh.path.Path` object.
    if hasattr(mesh, "entities"):
        if len(mesh.vertices) and mesh.vertices.shape[-1] != 3:
            raise ValueError("only Path3D export is supported for ply")

        if len(mesh.vertices) > 0:
            # run the discrete curve step for each entity
            discrete = [e.discrete(mesh.vertices) for e in mesh.entities]

            # how long was each discrete curve
            discrete_len = np.array([d.shape[0] for d in discrete])
            # what's the index offset based on these lengths
            discrete_off = np.concatenate(([0], np.cumsum(discrete_len)[:-1]))

            # pre-stack edges we can slice and offset
            longest = discrete_len.max()
            stack = np.column_stack((np.arange(0, longest - 1), np.arange(1, longest)))

            # get the indexes that reconstruct the discrete curves when stacked
            edges = np.vstack(
                [
                    stack[:length] + offset
                    for length, offset in zip(discrete_len - 1, discrete_off)
                ]
            )

            vertices = np.vstack(discrete)
            # create and populate the custom dtype for vertices
            num_vertices = len(vertices)
            # put mesh edge data into custom dtype to export
            num_edges = len(edges)

            if num_edges > 0 and num_vertices > 0:
                header.append(templates["vertex"])
                pack_vertex = np.zeros(num_vertices, dtype=dtype_vertex)
                pack_vertex["vertex"] = np.asarray(vertices, dtype=np.float32)

                # add the edge info to the header
                header.append(templates["edge"])
                # pack edges into our dtype
                pack_edges = unstructured_to_structured(edges, dtype=dtype_edge)

                # add the values for the header
                header_params.update(
                    {"edge_count": num_edges, "vertex_count": num_vertices}
                )

    elif hasattr(mesh, "vertices"):
        header.append(templates["vertex"])

        num_vertices = len(mesh.vertices)
        header_params["vertex_count"] = num_vertices
        # if we're exporting vertex normals add them
        # to the header and dtype
        if vertex_normal:
            header.append(templates["vertex_normal"])
            dtype_vertex.append(dtype_vertex_normal)

        # if mesh has a vertex color add it to the header
        vertex_color = (
            hasattr(mesh, "visual")
            and mesh.visual.kind == "vertex"
            and len(mesh.visual.vertex_colors) == len(mesh.vertices)
        )
        if vertex_color:
            header.append(templates["color"])
            dtype_vertex.append(dtype_color)

        if include_attributes:
            if hasattr(mesh, "vertex_attributes"):
                vertex_count = len(mesh.vertices)
                vertex_attributes = {
                    k: v
                    for k, v in mesh.vertex_attributes.items()
                    if hasattr(v, "__len__") and len(v) == vertex_count
                }
                _add_attributes_to_header(header, vertex_attributes)
                _add_attributes_to_dtype(dtype_vertex, vertex_attributes)
            else:
                vertex_attributes = None

        # create and populate the custom dtype for vertices
        pack_vertex = np.zeros(num_vertices, dtype=dtype_vertex)
        pack_vertex["vertex"] = mesh.vertices
        if vertex_normal:
            pack_vertex["normals"] = mesh.vertex_normals
        if vertex_color:
            pack_vertex["rgba"] = mesh.visual.vertex_colors

        if include_attributes and vertex_attributes is not None:
            _add_attributes_to_data_array(pack_vertex, vertex_attributes)

    if hasattr(mesh, "faces"):
        header.append(templates["face"])
        if mesh.visual.kind == "face" and encoding != "ascii":
            header.append(templates["color"])
            dtype_face.append(dtype_color)

        if include_attributes and hasattr(mesh, "face_attributes"):
            _add_attributes_to_header(header, mesh.face_attributes)
            _add_attributes_to_dtype(dtype_face, mesh.face_attributes)

        # put mesh face data into custom dtype to export
        pack_faces = np.zeros(len(mesh.faces), dtype=dtype_face)
        pack_faces["count"] = 3
        pack_faces["index"] = mesh.faces
        if mesh.visual.kind == "face" and encoding != "ascii":
            pack_faces["rgba"] = mesh.visual.face_colors
        header_params["face_count"] = len(mesh.faces)

        if include_attributes and hasattr(mesh, "face_attributes"):
            _add_attributes_to_data_array(pack_faces, mesh.face_attributes)

    header.append(templates["outro"])
    export = [Template("".join(header)).substitute(header_params).encode("utf-8")]

    if encoding == "binary_little_endian":
        if pack_vertex is not None:
            export.append(pack_vertex.tobytes())
        if pack_faces is not None:
            export.append(pack_faces.tobytes())
        if pack_edges is not None:
            export.append(pack_edges.tobytes())
    elif encoding == "ascii":
        if pack_vertex is not None:
            export.append(
                util.structured_array_to_string(
                    pack_vertex, col_delim=" ", row_delim="\n"
                ).encode("utf-8"),
            )

        if pack_faces is not None:
            export.extend(
                [
                    b"\n",
                    util.structured_array_to_string(
                        pack_faces, col_delim=" ", row_delim="\n"
                    ).encode("utf-8"),
                ]
            )

        if pack_edges is not None:
            export.extend(
                [
                    b"\n",
                    util.structured_array_to_string(
                        pack_edges, col_delim=" ", row_delim="\n"
                    ).encode("utf-8"),
                ]
            )
        export.append(b"\n")

    else:
        raise ValueError("encoding must be ascii or binary!")

    return b"".join(export)


def _parse_header(file_obj):
    """
    Read the ASCII header of a PLY file, and leave the file object
    at the position of the start of data but past the header.

    Parameters
    -----------
    file_obj : open file object
      Positioned at the start of the file

    Returns
    -----------
    elements : collections.OrderedDict
      Fields and data types populated
    is_ascii : bool
      Whether the data is ASCII or binary
    image_name : None or str
      File name of TextureFile
    """

    if "ply" not in str(file_obj.readline()).lower():
        raise ValueError("Not a ply file!")

    # collect the encoding: binary or ASCII
    encoding = file_obj.readline().decode("utf-8").strip().lower()
    is_ascii = "ascii" in encoding

    # big or little endian
    endian = ["<", ">"][int("big" in encoding)]
    elements = collections.OrderedDict()

    # store file name of TextureFiles in the header
    image_name = None

    while True:
        raw = file_obj.readline()
        if raw is None:
            raise ValueError("Header not terminated properly!")
        raw = raw.decode("utf-8").strip()
        line = raw.split()

        # we're done
        if "end_header" in line:
            break

        # elements are groups of properties
        if "element" in line[0]:
            # we got a new element so add it
            name, length = line[1:]
            elements[name] = {
                "length": int(length),
                "properties": collections.OrderedDict(),
            }
        # a property is a member of an element
        elif "property" in line[0]:
            # is the property a simple single value, like:
            # `property float x`
            if len(line) == 3:
                dtype, field = line[1:]
                elements[name]["properties"][str(field)] = endian + _dtypes[dtype]
            # is the property a painful list, like:
            # `property list uchar int vertex_indices`
            elif "list" in line[1]:
                dtype_count, dtype, field = line[2:]
                elements[name]["properties"][str(field)] = (
                    endian + _dtypes[dtype_count] + ", ($LIST,)" + endian + _dtypes[dtype]
                )
        # referenced as a file name
        elif "texturefile" in raw.lower():
            # textures come listed like:
            # `comment TextureFile fuze_uv.jpg`
            index = raw.lower().index("texturefile") + 11
            # use the value from raw to preserve whitespace
            image_name = raw[index:].strip()

    return elements, is_ascii, image_name


def _elements_to_kwargs(elements, fix_texture, image, prefer_color=None):
    """
    Given an elements data structure, extract the keyword
    arguments that a Trimesh object constructor will expect.

    Parameters
    ------------
    elements : OrderedDict object
      With fields and data loaded
    fix_texture : bool
      If True, will re- index vertices and faces
      so vertices with different UV coordinates
      are disconnected.
    image : PIL.Image
      Image to be viewed
    prefer_color : None, 'vertex', or 'face'
      Which kind of color to prefer if both defined

    Returns
    -----------
    kwargs : dict
      Keyword arguments for Trimesh constructor
    """
    # store the raw ply structure as an internal key in metadata
    kwargs = {"metadata": {"_ply_raw": elements}}

    if "vertex" in elements and elements["vertex"]["length"]:
        vertices = np.column_stack([elements["vertex"]["data"][i] for i in "xyz"])
        if not util.is_shape(vertices, (-1, 3)):
            raise ValueError("Vertices were not (n,3)!")
    else:
        # return empty geometry if there are no vertices
        kwargs["geometry"] = {}
        return kwargs

    try:
        vertex_normals = np.column_stack(
            [elements["vertex"]["data"][j] for j in ("nx", "ny", "nz")]
        )
        if len(vertex_normals) == len(vertices):
            kwargs["vertex_normals"] = vertex_normals
    except BaseException:
        pass

    if "face" in elements and elements["face"]["length"]:
        face_data = elements["face"]["data"]
    else:
        # some PLY files only include vertices
        face_data = None
        faces = None

    # what keys do in-the-wild exporters use for vertices
    index_names = ["vertex_index", "vertex_indices"]
    texcoord = None

    if util.is_shape(face_data, (-1, (3, 4))):
        faces = face_data
    elif isinstance(face_data, dict):
        # get vertex indexes
        for i in index_names:
            if i in face_data:
                faces = face_data[i]
                break
        # if faces have UV coordinates defined use them
        if "texcoord" in face_data:
            texcoord = face_data["texcoord"]

    elif isinstance(face_data, np.ndarray):
        face_blob = elements["face"]["data"]
        # some exporters set this name to 'vertex_index'
        # and some others use 'vertex_indices' but we really
        # don't care about the name unless there are multiple
        if len(face_blob.dtype.names) == 1:
            name = face_blob.dtype.names[0]
        elif len(face_blob.dtype.names) > 1:
            # loop through options
            for i in face_blob.dtype.names:
                if i in index_names:
                    name = i
                    break
        # get faces
        faces = face_blob[name]["f1"]

        try:
            texcoord = face_blob["texcoord"]["f1"]
        except (ValueError, KeyError):
            # accessing numpy arrays with named fields
            # incorrectly is a ValueError
            pass

    if faces is not None:
        shape = np.shape(faces)
        if len(shape) != 2:
            # we may have mixed quads and triangles handle them with function
            faces = triangulate_quads(faces)

        if texcoord is None:
            # ply has no clear definition of how texture coordinates are stored,
            # unfortunately there are many common names that we need to try
            texcoord_names = [("texture_u", "texture_v"), ("u", "v"), ("s", "t")]
            for names in texcoord_names:
                # If texture coordinates are defined with vertices
                try:
                    t_u = elements["vertex"]["data"][names[0]]
                    t_v = elements["vertex"]["data"][names[1]]
                    texcoord = np.stack(
                        (t_u[faces.reshape(-1)], t_v[faces.reshape(-1)]), axis=-1
                    ).reshape((faces.shape[0], -1))
                    # stop trying once succeeded
                    break
                except (ValueError, KeyError):
                    # if the fields didn't exist
                    pass

        shape = np.shape(faces)

        # PLY stores texture coordinates per-face which is
        # slightly annoying, as we have to then figure out
        # which vertices have the same position but different UV
        if (
            texcoord is not None
            and len(shape) == 2
            and texcoord.shape == (faces.shape[0], faces.shape[1] * 2)
        ):
            # vertices with the same position but different
            # UV coordinates can't be merged without it
            # looking like it went through a woodchipper
            # in- the- wild PLY comes with things merged that
            # probably shouldn't be so disconnect vertices
            if fix_texture:
                # do import here
                from ..visual.texture import unmerge_faces

                # reshape to correspond with flattened faces
                uv_all = texcoord.reshape((-1, 2))
                # UV coordinates defined for every triangle have
                # duplicates which can be merged so figure out
                # which UV coordinates are the same here
                unique, inverse = grouping.unique_rows(uv_all)

                # use the indices of faces and face textures
                # to only merge vertices where the position
                # AND uv coordinate are the same
                faces, mask_v, mask_vt = unmerge_faces(
                    faces, inverse.reshape(faces.shape)
                )
                # apply the mask to get resulting vertices
                vertices = vertices[mask_v]
                # apply the mask to get UV coordinates
                uv = uv_all[unique][mask_vt]
            else:
                # don't alter vertices, UV will look like crap
                # if it was exported with vertices merged
                uv = np.zeros((len(vertices), 2))
                uv[faces.reshape(-1)] = texcoord.reshape((-1, 2))

            # create the visuals object for the texture
            kwargs["visual"] = visual.texture.TextureVisuals(uv=uv, image=image)
        elif texcoord is not None:
            # create a texture with an empty material
            from ..visual.texture import TextureVisuals

            uv = np.zeros((len(vertices), 2))
            uv[faces.reshape(-1)] = texcoord.reshape((-1, 2))
            kwargs["visual"] = TextureVisuals(uv=uv)
        # faces were not none so assign them
        kwargs["faces"] = faces
    # kwargs for Trimesh or PointCloud
    kwargs["vertices"] = vertices

    # if both vertex and face color are defined pick the one
    if "face" in elements:
        kwargs["face_colors"] = _element_colors(elements["face"])
    if "vertex" in elements:
        kwargs["vertex_colors"] = _element_colors(elements["vertex"])

    if "edge" in elements:
        # this is a Path element.

        from ..path.exchange.misc import edges_to_path

        edges = structured_to_unstructured(elements["edge"]["data"])
        kwargs.update(edges_to_path(edges, kwargs["vertices"]))

    return kwargs


def _element_colors(element):
    """
    Given an element, try to extract RGBA color from
    properties and return them as an (n,3|4) array.

    Parameters
    -------------
    element : dict
      Containing color keys

    Returns
    ------------
    colors : (n, 3) or (n, 4) float
      Colors extracted from the element
    signal : float
      Estimate of range
    """
    keys = ["red", "green", "blue", "alpha"]
    candidate_colors = [element["data"][i] for i in keys if i in element["properties"]]
    if len(candidate_colors) >= 3:
        return np.column_stack(candidate_colors)
    return None


def _load_element_different(properties, data):
    """
    Load elements which include lists of different lengths
    based on the element's property-definitions.

    Parameters
    ------------
    properties : dict
      Property definitions encoded in a dict where the property name is the key
      and the property data type the value.
    data : array
      Data rows for this element.
    """
    edata = {k: [] for k in properties.keys()}
    for row in data:
        start = 0
        for name, dt in properties.items():
            length = 1
            if "$LIST" in dt:
                dt = dt.split("($LIST,)")[-1]
                # the first entry in a list-property is the number of elements
                # in the list
                length = int(row[start])
                # skip the first entry (the length), when reading the data
                start += 1
            end = start + length
            edata[name].append(row[start:end].astype(dt))
            # start next property at the end of this one
            start = end

    # if the shape of any array is (n, 1) we want to
    # squeeze/concatenate it into (n,)
    squeeze = {k: np.array(v, dtype="object") for k, v in edata.items()}
    # squeeze and convert any clean 2D arrays
    squeeze.update(
        {
            k: v.squeeze().astype(edata[k][0].dtype)
            for k, v in squeeze.items()
            if len(v.shape) == 2
        }
    )

    return squeeze


def _load_element_single(properties, data):
    """
    Load element data with lists of a single length
    based on the element's property-definitions.

    Parameters
    ------------
    properties : dict
      Property definitions encoded in a dict where
      the property name is the key and the property
      data type the value.
    data : array
      Data rows for this element, if the data contains
      list-properties all lists belonging to one property
      must have the same length.
    """

    first = data[0]
    columns = {}
    current = 0
    for name, dt in properties.items():
        # if the current index has gone past the number
        # of items we actually have exit the loop early
        if current >= len(first):
            break
        if "$LIST" in dt:
            dtype = dt.split("($LIST,)")[-1]
            # the first entry in a list-property
            # is the number of elements in the list

            length = int(first[current])
            columns[name] = data[:, current + 1 : current + 1 + length].astype(dtype)
            # offset by length of array plus one for each uint index
            current += length + 1
        else:
            columns[name] = data[:, current : current + 1].astype(dt)
            current += 1

    return columns


def _ply_ascii(elements, file_obj):
    """
    Load data from an ASCII PLY file into an existing elements data structure.

    Parameters
    ------------
    elements : OrderedDict
      Populated from the file header, data will
      be added in-place to this object
    file_obj : file-like-object
      Current position at the start
      of the data section (past the header).
    """

    # get the file contents as a string
    text = str(file_obj.read().decode("utf-8"))
    # split by newlines
    lines = str.splitlines(text)
    # get each line as an array split by whitespace
    array = [np.fromstring(i, sep=" ") for i in lines]
    # store the line position in the file
    row_pos = 0

    # loop through data we need
    for key, values in elements.items():
        # if the element is empty ignore it
        if "length" not in values or values["length"] == 0:
            continue
        data = array[row_pos : row_pos + values["length"]]
        row_pos += values["length"]
        # try stacking the data, which simplifies column-wise access. this is only
        # possible, if all rows have the same length.
        try:
            data = np.vstack(data)
            col_count_equal = True
        except ValueError:
            col_count_equal = False

        # number of list properties in this element
        list_count = sum(1 for dt in values["properties"].values() if "$LIST" in dt)
        if col_count_equal and list_count <= 1:
            # all rows have the same length and we only have at most one list
            # property where all entries have the same length. this means we can
            # use the quick numpy-based loading.
            element_data = _load_element_single(values["properties"], data)
        else:
            # there are lists of differing lengths. we need to fall back to loading
            # the data by iterating all rows and checking for list-lengths. this is
            # slower than the variant above.
            element_data = _load_element_different(values["properties"], data)

        elements[key]["data"] = element_data


def _ply_binary(elements, file_obj):
    """
    Load the data from a binary PLY file into the elements data structure.

    Parameters
    ------------
    elements : OrderedDict
      Populated from the file header.
      Object will be modified to add data by this function.
    file_obj : open file object
      With current position at the start
      of the data section (past the header)
    """

    def populate_listsize(file_obj, elements):
        """
        Given a set of elements populated from the header if there are any
        list properties seek in the file the length of the list.

        Note that if you have a list where each instance is different length
        (if for example you mixed triangles and quads) this won't work at all
        """
        p_start = file_obj.tell()
        p_current = file_obj.tell()
        elem_pop = []
        for element_key, element in elements.items():
            props = element["properties"]
            prior_data = ""
            for k, dtype in props.items():
                prop_pop = []
                if "$LIST" in dtype:
                    # every list field has two data types:
                    # the list length (single value), and the list data (multiple)
                    # here we are only reading the single value for list length
                    field_dtype = np.dtype(dtype.split(",")[0])
                    if len(prior_data) == 0:
                        offset = 0
                    else:
                        offset = np.dtype(prior_data).itemsize
                    file_obj.seek(p_current + offset)
                    blob = file_obj.read(field_dtype.itemsize)
                    if len(blob) == 0:
                        # no data was read for property
                        prop_pop.append(k)
                        break
                    size = np.frombuffer(blob, dtype=field_dtype)[0]
                    props[k] = props[k].replace("$LIST", str(size))
                prior_data += props[k] + ","
            if len(prop_pop) > 0:
                # if a property was empty remove it
                for pop in prop_pop:
                    props.pop(pop)
                # if we've removed all properties from
                # an element remove the element later
                if len(props) == 0:
                    elem_pop.append(element_key)
                    continue
            # get the size of the items in bytes
            itemsize = np.dtype(", ".join(props.values())).itemsize
            # offset the file based on read size
            p_current += element["length"] * itemsize
        # move the file back to where we found it
        file_obj.seek(p_start)
        # if there were elements without properties remove them
        for pop in elem_pop:
            elements.pop(pop)

    def populate_data(file_obj, elements):
        """
        Given the data type and field information from the header,
        read the data and add it to a 'data' field in the element.
        """
        for key in elements.keys():
            items = list(elements[key]["properties"].items())
            dtype = np.dtype(items)
            data = file_obj.read(elements[key]["length"] * dtype.itemsize)
            try:
                elements[key]["data"] = np.frombuffer(data, dtype=dtype)
            except BaseException:
                log.warning(f"PLY failed to populate: {key}")
                elements[key]["data"] = None
        return elements

    def _elements_size(elements):
        """
        Given an elements data structure populated from the header,
        calculate how long the file should be if it is intact.
        """
        size = 0
        for element in elements.values():
            dtype = np.dtype(",".join(element["properties"].values()))
            size += element["length"] * dtype.itemsize
        return size

    # some elements are passed where the list dimensions
    # are not included in the header, so this function goes
    # into the meat of the file and grabs the list dimensions
    # before we to the main data read as a single operation
    populate_listsize(file_obj, elements)

    # how many bytes are left in the file
    size_file = util.distance_to_end(file_obj)
    # how many bytes should the data structure described by
    # the header take up
    size_elements = _elements_size(elements)

    # if the number of bytes is not the same the file is probably corrupt
    if size_file != size_elements:
        raise ValueError("PLY is unexpected length!")

    # with everything populated and a reasonable confidence the file
    # is intact, read the data fields described by the header
    populate_data(file_obj, elements)


def export_draco(mesh, bits=28):
    """
    Export a mesh using Google's Draco compressed format.

    Only works if draco_encoder is in your PATH:
    https://github.com/google/draco

    Parameters
    ----------
    mesh : Trimesh object
      Mesh to export
    bits : int
      Bits of quantization for position
      tol.merge=1e-8 is roughly 25 bits

    Returns
    ----------
    data : str or bytes
      DRC file bytes
    """
    with tempfile.NamedTemporaryFile(suffix=".ply") as temp_ply:
        temp_ply.write(export_ply(mesh))
        temp_ply.flush()
        with tempfile.NamedTemporaryFile(suffix=".drc") as encoded:
            subprocess.check_output(
                [
                    draco_encoder,
                    "-qp",
                    str(int(bits)),
                    "-i",
                    temp_ply.name,
                    "-o",
                    encoded.name,
                ]
            )
            encoded.seek(0)
            data = encoded.read()
    return data


def load_draco(file_obj, **kwargs):
    """
    Load a mesh from Google's Draco format.

    Parameters
    ----------
    file_obj : file- like object
      Contains data

    Returns
    ----------
    kwargs : dict
      Keyword arguments to construct a Trimesh object
    """

    with tempfile.NamedTemporaryFile(suffix=".drc") as temp_drc:
        temp_drc.write(file_obj.read())
        temp_drc.flush()

        with tempfile.NamedTemporaryFile(suffix=".ply") as temp_ply:
            subprocess.check_output(
                [draco_decoder, "-i", temp_drc.name, "-o", temp_ply.name]
            )
            temp_ply.seek(0)
            kwargs = load_ply(temp_ply)
    return kwargs


_ply_loaders = {"ply": load_ply}
_ply_exporters = {"ply": export_ply}

draco_encoder = util.which("draco_encoder")
draco_decoder = util.which("draco_decoder")
if draco_decoder is not None:
    _ply_loaders["drc"] = load_draco
if draco_encoder is not None:
    _ply_exporters["drc"] = export_draco



try:
    from . import generic as g
except BaseException:
    import generic as g


class CurvatureTest(g.unittest.TestCase):
    def test_gaussian_curvature(self):
        for radius in g.np.linspace(0.25, 2.0, 10):
            m = g.trimesh.creation.icosphere(radius=radius)

            gauss = g.trimesh.curvature.discrete_gaussian_curvature_measure(
                mesh=m, points=m.vertices, radius=radius * 2.0
            ) / (4 * g.np.pi)
            assert g.np.allclose(gauss, 1.0, atol=0.01)

        # a torus should have approximately half its vertices with positive
        # curvature, and half with negative
        t = g.get_mesh("torus.STL")
        gauss = g.trimesh.curvature.discrete_gaussian_curvature_measure(
            mesh=t, points=t.vertices, radius=1.0
        )
        ratio = float((gauss < 0.0).sum()) / float(len(gauss))
        assert g.np.isclose(ratio, 0.5, atol=0.2)

    def test_mean_curvature(self):
        m = g.trimesh.creation.icosphere()
        mean = g.trimesh.curvature.discrete_mean_curvature_measure(m, m.vertices, 2.0) / (
            4 * g.np.pi
        )
        # Check mean curvature for unit sphere is 1.0
        assert g.np.allclose(mean, 1.0, atol=0.01)

    def test_vertex_defect(self):
        # a subdivided box will only have corners and planar regions
        # so all vertex defects should be 0 or 90 degrees
        m = g.trimesh.primitives.Box().subdivide()
        assert g.np.logical_or(
            g.np.isclose(m.vertex_defects, 0.0),
            g.np.isclose(m.vertex_defects, g.np.pi / 2.0),
        ).all()

        assert len(m.vertex_defects) == len(m.vertices)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



import os

import numpy as np

from ..constants import log, tol
from ..version import __version__


def export_urdf(mesh, directory, scale=1.0, color=None, **kwargs):
    """
    Convert a Trimesh object into a URDF package for physics
    simulation. This breaks the mesh into convex pieces and
    writes them to the same directory as the .urdf file.

    Parameters
    ---------
    mesh : trimesh.Trimesh
      Input geometry
    directory : str
      The directory path for the URDF package

    Returns
    ---------
    mesh : Trimesh
      Multi-body mesh containing convex decomposition
    """

    import lxml.etree as et

    # TODO: fix circular import
    from .export import export_mesh

    # Extract the save directory and the file name
    fullpath = os.path.abspath(directory)
    name = os.path.basename(fullpath)
    _, ext = os.path.splitext(name)

    if ext != "":
        raise ValueError("URDF path must be a directory!")

    # Create directory if needed
    if not os.path.exists(fullpath):
        os.mkdir(fullpath)
    elif not os.path.isdir(fullpath):
        raise ValueError("URDF path must be a directory!")

    # Perform a convex decomposition
    try:
        convex_pieces = mesh.convex_decomposition()
    except BaseException:
        log.error("problem with convex decomposition, using hull", exc_info=True)
        convex_pieces = [mesh.convex_hull]

    # Get the effective density of the mesh
    effective_density = mesh.volume / sum([m.volume for m in convex_pieces])

    # open an XML tree
    root = et.Element("robot", name="root")

    # Loop through all pieces, adding each as a link
    prev_link_name = None
    for i, piece in enumerate(convex_pieces):
        # Save each nearly convex mesh out to a file
        piece_name = f"{name}_convex_piece_{i}"
        piece_filename = f"{piece_name}.obj"
        piece_filepath = os.path.join(fullpath, piece_filename)
        export_mesh(piece, piece_filepath)

        # Set the mass properties of the piece
        piece.center_mass = mesh.center_mass
        piece.density = effective_density * mesh.density

        link_name = f"link_{piece_name}"
        geom_name = f"{piece_filename}"
        I = [["{:.2E}".format(y) for y in x] for x in piece.moment_inertia]  # NOQA

        # Write the link out to the XML Tree
        link = et.SubElement(root, "link", name=link_name)

        # Inertial information
        inertial = et.SubElement(link, "inertial")
        et.SubElement(inertial, "origin", xyz="0 0 0", rpy="0 0 0")
        et.SubElement(inertial, "mass", value=f"{piece.mass:.2E}")
        et.SubElement(
            inertial,
            "inertia",
            ixx=I[0][0],
            ixy=I[0][1],
            ixz=I[0][2],
            iyy=I[1][1],
            iyz=I[1][2],
            izz=I[2][2],
        )
        # Visual Information
        visual = et.SubElement(link, "visual")
        et.SubElement(visual, "origin", xyz="0 0 0", rpy="0 0 0")
        geometry = et.SubElement(visual, "geometry")
        et.SubElement(
            geometry,
            "mesh",
            filename=geom_name,
            scale=f"{scale:.4E} {scale:.4E} {scale:.4E}",
        )
        material = et.SubElement(visual, "material", name="")
        if color is not None:
            et.SubElement(
                material, "color", rgba=f"{color[0]:.2E} {color[1]:.2E} {color[2]:.2E} 1"
            )

        # Collision Information
        collision = et.SubElement(link, "collision")
        et.SubElement(collision, "origin", xyz="0 0 0", rpy="0 0 0")
        geometry = et.SubElement(collision, "geometry")
        et.SubElement(
            geometry,
            "mesh",
            filename=geom_name,
            scale=f"{scale:.4E} {scale:.4E} {scale:.4E}",
        )

        # Create rigid joint to previous link
        if prev_link_name is not None:
            joint_name = f"{link_name}_joint"
            joint = et.SubElement(root, "joint", name=joint_name, type="fixed")
            et.SubElement(joint, "origin", xyz="0 0 0", rpy="0 0 0")
            et.SubElement(joint, "parent", link=prev_link_name)
            et.SubElement(joint, "child", link=link_name)

        prev_link_name = link_name

    # Write URDF file
    tree = et.ElementTree(root)
    urdf_filename = f"{name}.urdf"
    tree.write(os.path.join(fullpath, urdf_filename), pretty_print=True)

    # Write Gazebo config file
    root = et.Element("model")
    model = et.SubElement(root, "name")
    model.text = name
    version = et.SubElement(root, "version")
    version.text = "1.0"
    sdf = et.SubElement(root, "sdf", version="1.4")
    sdf.text = f"{name}.urdf"

    author = et.SubElement(root, "author")
    et.SubElement(author, "name").text = f"trimesh {__version__}"
    et.SubElement(author, "email").text = "blank@blank.blank"

    description = et.SubElement(root, "description")
    description.text = name
    tree = et.ElementTree(root)

    if tol.strict:
        from ..resources import get_stream

        # todo : we don't pass the URDF schema validation
        schema = et.XMLSchema(file=get_stream("schema/urdf.xsd"))
        if not schema.validate(tree):
            # actual error isn't raised by validate
            log.debug(schema.error_log)

    tree.write(os.path.join(fullpath, "model.config"))
    return np.sum(convex_pieces)



"""
Load all the meshes we can get our hands on and check things, stuff.
"""

try:
    from . import generic as g
except BaseException:
    import generic as g


def test_meshes():
    # make sure we can load everything we think we can
    formats = g.trimesh.available_formats()
    assert all(isinstance(i, str) for i in formats)
    assert all(len(i) > 0 for i in formats)
    assert all(i in formats for i in ["stl", "ply", "off", "obj", "glb"])

    for mesh in g.get_meshes(raise_error=True):
        # log file name for debugging
        file_name = mesh.source.file_name

        # ply files can return PointCloud objects
        if file_name.startswith("points_"):
            continue

        g.log.info("Testing %s", file_name)
        start = {mesh.__hash__(), mesh.__hash__()}
        assert len(mesh.faces) > 0
        assert len(mesh.vertices) > 0

        # make sure vertex normals match vertices and are valid
        assert mesh.vertex_normals.shape == mesh.vertices.shape
        assert g.np.isfinite(mesh.vertex_normals).all()

        # should be one per vertex
        assert len(mesh.vertex_faces) == len(mesh.vertices)

        # check some edge properties
        assert len(mesh.edges) > 0
        assert len(mesh.edges_unique) > 0
        assert len(mesh.edges_sorted) == len(mesh.edges)
        assert len(mesh.edges_face) == len(mesh.edges)

        # check edges_unique
        assert len(mesh.edges) == len(mesh.edges_unique_inverse)
        assert g.np.allclose(
            mesh.edges_sorted, mesh.edges_unique[mesh.edges_unique_inverse]
        )
        assert len(mesh.edges_unique) == len(mesh.edges_unique_length)

        # euler number should be an integer
        assert isinstance(mesh.euler_number, int)

        # check bounding primitives
        assert mesh.bounding_box.volume > 0.0
        assert mesh.bounding_primitive.volume > 0.0

        # none of these should have mutated anything
        assert start == {mesh.__hash__(), mesh.__hash__()}

        # run processing, again
        mesh.process()

        # still shouldn't have changed anything
        assert start == {mesh.__hash__(), mesh.__hash__()}

        if not (mesh.is_watertight and mesh.is_winding_consistent):
            continue

        assert len(mesh.facets) == len(mesh.facets_area)
        assert len(mesh.facets) == len(mesh.facets_normal)
        assert len(mesh.facets) == len(mesh.facets_boundary)

        if len(mesh.facets) != 0:
            faces = mesh.facets[mesh.facets_area.argmax()]
            outline = mesh.outline(faces)
            # check to make sure we can generate closed paths
            # on a Path3D object
            test = outline.paths  # NOQA

        smoothed = mesh.smooth_shaded  # NOQA

        assert abs(mesh.volume) > 0.0

        mesh.section(plane_normal=[0, 0, 1], plane_origin=mesh.centroid)

        sample = mesh.sample(1000)
        even_sample = g.trimesh.sample.sample_surface_even(mesh, 100)  # NOQA
        assert sample.shape == (1000, 3)
        g.log.info("finished testing meshes")

        # make sure vertex kdtree and triangles rtree exist

        t = mesh.kdtree
        assert hasattr(t, "query")
        g.log.info("Creating triangles tree")
        r = mesh.triangles_tree
        assert hasattr(r, "intersection")
        g.log.info("Triangles tree ok")

        # face angles should have same
        assert mesh.face_angles.shape == mesh.faces.shape
        assert len(mesh.vertices) == len(mesh.vertex_defects)
        assert len(mesh.principal_inertia_components) == 3

        # make a ray query which may lead to unpicklable caching
        dimension = (100, 3)
        ray_origins = g.random(dimension)
        ray_directions = g.np.tile([0, 0, 1], (dimension[0], 1))
        ray_origins[:, 2] = mesh.bounds[0][2] - mesh.scale

        # call additional C objects
        assert mesh.kdtree is not None
        assert mesh.triangles_tree is not None

        # force ray object to be created
        ray = mesh.ray.intersects_location(ray_origins, ray_directions)
        assert ray is not None

        # collect list of cached properties that are writeable
        writeable = []

        # make sure a roundtrip pickle works
        # if the cache has non-pickleable stuff this will break
        pickle = g.pickle.dumps(mesh)
        assert isinstance(pickle, bytes)
        assert len(pickle) > 0

        r = g.pickle.loads(pickle)
        assert r.faces.shape == mesh.faces.shape
        assert g.np.isclose(r.volume, mesh.volume)

        # we should have built up a bunch of stuff into
        # our cache, so make sure all numpy arrays cached
        # are read-only and not crazy
        for name, cached in mesh._cache.cache.items():
            # only check numpy arrays
            if not isinstance(cached, g.np.ndarray):
                continue

            # nothing in the cache should be writeable
            if cached.flags["WRITEABLE"]:
                raise ValueError(f"{name} is writeable!")

            # only check int, float, and bool
            if cached.dtype.kind not in "ibf":
                continue

            # there should never be NaN values
            if g.np.isnan(cached).any():
                raise ValueError("NaN values in %s/%s", file_name, name)

            # fields allowed to have infinite values
            if name in ["face_adjacency_radius"]:
                continue

            # make sure everything is finite
            if not g.np.isfinite(cached).all():
                raise ValueError("inf values in %s/%s", file_name, name)

        # ...still shouldn't have changed anything
        assert start == {mesh.__hash__(), mesh.__hash__()}

        # log the names of properties we need to make read-only
        if len(writeable) > 0:
            # TODO : all cached values should be read-only
            g.log.error("cached properties writeable: {}".format(", ".join(writeable)))


def test_mesh_2D():
    # check a simple mesh with 2D vertices
    m = g.trimesh.Trimesh(
        vertices=g.np.random.random((100, 2)),
        faces=g.np.arange(99, dtype=g.np.int64).reshape((-1, 3)),
    )
    # the face normals should be 3D (+Z)
    assert m.face_normals.shape == m.faces.shape

    rend = g.trimesh.rendering.mesh_to_vertexlist(m)
    assert len(rend) > 1


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    test_mesh_2D()



"""
Copy meshes and make sure they do what we expect.
"""

try:
    from . import generic as g
except BaseException:
    import generic as g


class CopyTests(g.unittest.TestCase):
    def test_copy(self):
        for mesh in g.get_meshes(raise_error=True):
            if not isinstance(mesh, g.trimesh.Trimesh) or len(mesh.faces) == 0:
                continue
            start = hash(mesh)

            # make sure some stuff is populated
            _ = mesh.kdtree
            _ = mesh.triangles_tree
            _ = mesh.face_adjacency_angles
            _ = mesh.facets
            assert "triangles_tree" in mesh._cache
            assert len(mesh._cache) > 0

            # if you cache c-objects then deepcopy the mesh
            # it randomly segfaults
            copy_count = 200
            for _i in range(copy_count):
                copied = mesh.copy(include_cache=False)
                assert len(copied._cache) == 0
                assert len(mesh._cache) > 0

                # deepcopy should clear the cache
                copied = g.copy.deepcopy(mesh)
                assert len(copied._cache) == 0
                assert len(mesh._cache) > 0

                # regular copy should try to preserve the cache
                copied = g.copy.copy(mesh)
                assert len(copied._cache) == len(mesh._cache)
                # the triangles_tree should be the SAME OBJECT
                assert id(copied.triangles_tree) == id(mesh.triangles_tree)

                # cache should be same data in different object
                assert id(copied._cache.cache) != id(mesh._cache.cache)
                assert id(copied._cache) != id(mesh._cache)
                # identifier shouldn't change
                assert g.np.allclose(copied.identifier, mesh.identifier)

            # ...still shouldn't have changed anything
            assert start == hash(mesh)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()
